# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import pytest

from portmod.loader import load_installed_pkg
from portmod.merge import configure
from portmodlib.atom import Atom

from .env import setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup():
    """
    Sets up and tears down the test environment
    """
    dictionary = setup_env("test")
    yield dictionary
    tear_down_env()


def test_nodeps_use_dependency(setup):
    """Tests that simple dependency resolution works"""
    configure(["=test/test7-1.0[baz]"], nodeps=True)
    # A dependency which would otherwise be pulled in by baz, but nodeps should ignore it
    assert not load_installed_pkg(Atom("test/test5"))
    pkg = load_installed_pkg(Atom("test/test7"))
    assert pkg
    assert "baz" in pkg.INSTALLED_USE



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import json
import lzma
import os
import shutil
from logging import error, info, warning
from typing import Optional, Set

from portmod.cfg_protect import get_protected_path, is_protected
from portmod.config import get_config_value, variable_data_dir
from portmod.download import download_mod
from portmod.globals import env
from portmod.loader import (
    _delete_state,
    _sandbox_execute_pybuild,
    _state_path,
    load_all_installed,
    load_installed_pkg,
)
from portmod.parsers.manifest import FileType, Manifest, ManifestEntry
from portmod.vdb import VDB, vdb_path
from portmodlib._phase import PhaseState
from portmodlib.atom import Atom
from portmodlib.colour import green
from portmodlib.fs import (
    _iter_files_to_patch,
    _patch_file,
    ci_exists,
    get_tree_size,
    is_parent,
    onerror,
)
from portmodlib.l10n import l10n
from portmodlib.module_util import _add_redirection

from ._deprecated.rebuild import get_rebuild_manifest
from ._deprecated.vfs import _cleanup_tmp_archive_dir
from .cache import clear_cache_for_path
from .functools import clear_install_cache
from .modules import module_prerm
from .perms import Permissions
from .prompt import prompt_options
from .pybuild import InstalledPybuild, Pybuild


def belongs(path: str, case_sensitive: bool = True) -> Optional[InstalledPybuild]:
    """
    Returns the package which installed the given file

    args:
        The path of the file to check, relative to the installation root

    If no such package exists, returns None
    """
    normpath = os.path.normpath(path)
    for pkg in load_all_installed():
        if pkg.get_contents().get(normpath, case_sensitive=case_sensitive):
            return pkg

    return None


def src_unpack(
    pkg: Pybuild,
    build_dir: str,
    curdir: Optional[str] = None,
    state: PhaseState = PhaseState(),
):
    permissions = Permissions(
        rw_paths=[build_dir],
        ro_paths=[env.DOWNLOAD_DIR],
        global_read=False,
        network=True,
        tmp=state.T,
    )
    _sandbox_execute_pybuild(
        pkg.FILE,
        "unpack",
        permissions,
        save_state=True,
        init=state.__dict__,
        curdir=curdir or build_dir,
    )


def src_prepare(
    pkg: Pybuild,
    build_dir: str,
    curdir: Optional[str] = None,
    state: PhaseState = PhaseState(),
):
    # Default does nothing unless pkg.PATCHES is set
    if pkg.PATCHES or "src_prepare" in pkg.FUNCTIONS:
        permissions = Permissions(
            rw_paths=[build_dir], global_read=True, network=False, tmp=state.T
        )
        _sandbox_execute_pybuild(
            pkg.FILE,
            "prepare",
            permissions,
            save_state=True,
            init=state.__dict__,
            curdir=curdir or build_dir,
        )


def src_install(
    pkg: Pybuild,
    build_dir: str,
    curdir: Optional[str] = None,
    state: PhaseState = PhaseState(),
):
    permissions = Permissions(
        rw_paths=[build_dir], global_read=True, network=False, tmp=state.T
    )
    _sandbox_execute_pybuild(
        pkg.FILE,
        "install",
        permissions,
        save_state=True,
        init=state.__dict__,
        curdir=curdir or build_dir,
    )


def pkg_postinst(
    pkg: Pybuild, final_install: str, curdir: str, state: PhaseState = PhaseState()
):
    # Default does nothing
    if "pkg_postinst" in pkg.FUNCTIONS:
        rw_paths = []
        if pkg._PYBUILD_VER == 1:
            rw_paths.append(final_install)

        permissions = Permissions(
            rw_paths=rw_paths,
            global_read=True,
            network=False,
            tmp=state.T,
        )
        _sandbox_execute_pybuild(
            pkg.FILE,
            "postinst",
            permissions,
            save_state=True,
            init=state.__dict__,
            curdir=curdir,
        )


def pkg_prerm(pkg: InstalledPybuild, root: str, state: PhaseState = PhaseState()):
    # Default does nothing
    if "pkg_prerm" in pkg.FUNCTIONS:
        rw_paths = []
        if pkg._PYBUILD_VER == 1:
            rw_paths.append(root)
        permissions = Permissions(
            rw_paths=rw_paths, global_read=True, network=False, tmp=state.T
        )
        _sandbox_execute_pybuild(
            pkg.FILE,
            "prerm",
            permissions,
            save_state=False,
            init=state.__dict__,
            curdir=root,
        )
    for path, _ in pkg.get_contents().entries.items():
        if path.endswith(".pmodule"):
            try:
                module_prerm(os.path.join(root, path))
            except Exception as e:
                error(e)


def remove_pkg(pkg: InstalledPybuild, reinstall: bool = False):
    """
    Removes the given mod

    args:
        reinstall: if true, don't touch the installed DB since we'll
                   need it to finish the install
    """
    print(">>> " + l10n("pkg-removing", atom=green(pkg.ATOM.CPF)))

    BUILD_DIR = os.path.join(env.TMP_DIR, pkg.CATEGORY, pkg.P)
    state = PhaseState(BUILD_DIR)
    assert state.T

    state.USE = pkg.INSTALLED_USE
    os.makedirs(state.T, exist_ok=True)

    state.ROOT = env.prefix().ROOT
    pkg_prerm(pkg, state.ROOT, state)
    del state.ROOT

    for path, entry in pkg.get_contents().entries.items():
        # If file on disk matches the one in the manifest, remove it.
        # Otherwise, warn the user and leave it as-is
        fullpath = os.path.join(env.prefix().ROOT, path)
        if os.path.exists(fullpath):
            realentry = ManifestEntry.from_path(FileType.MISC, fullpath, path)
            if realentry == entry:
                os.remove(fullpath)
                if os.path.exists(fullpath + ".orig"):
                    info(l10n("restoring-file-backup", path=fullpath + ".orig"))
                    os.rename(fullpath + ".orig", fullpath)
                # Remove the parent directory. If it is non-empty, this will fail.
                try:
                    os.removedirs(os.path.dirname(fullpath))
                except OSError:
                    pass
            else:
                warning(l10n("package-remove-file-conflict", path=path))

    db_path = os.path.join(vdb_path(), pkg.CATEGORY, pkg.PN)
    if os.path.exists(db_path) and not reinstall:
        with VDB() as vdb:
            # Remove and stage changes
            vdb.git.rm(os.path.join(pkg.CATEGORY, pkg.PN), r=True, f=True)
            # Clean up unstaged files (e.g. pycache)
            if os.path.exists(db_path):
                shutil.rmtree(db_path, onerror=onerror)
            clear_cache_for_path(os.path.join(db_path, os.path.basename(pkg.FILE)))

    # Remove from pybuild cache
    path = os.path.join(env.prefix().PYBUILD_INSTALLED_CACHE, pkg.CATEGORY, pkg.PF)
    if os.path.exists(path):
        os.remove(path)

    # Cleanup archive dir in case vfs had to extract anything
    _cleanup_tmp_archive_dir()

    print(">>> " + l10n("pkg-finished-removing", atom=green(pkg.ATOM.CPF)))
    clear_install_cache()


def install_pkg(mod: Pybuild, use_flags: Set[str]):
    print(">>> " + l10n("pkg-installing", atom=green(mod.ATOM.CPF)))
    old_curdir = os.getcwd()
    sources = download_mod(mod, use_flags)
    if sources is None:
        error(">>> " + l10n("pkg-unable-to-download", atom=green(mod.ATOM.CPF)))
        return False

    BUILD_DIR = os.path.join(env.TMP_DIR, mod.CATEGORY, mod.P)
    if os.path.exists(BUILD_DIR):
        shutil.rmtree(BUILD_DIR, onerror=onerror)
    state = PhaseState(BUILD_DIR)
    assert state.T
    state.A = [source.as_source() for source in sources]
    state.USE = use_flags

    # Ensure build directory is clean
    if os.path.exists(BUILD_DIR):
        shutil.rmtree(BUILD_DIR, onerror=onerror)

    state.WORKDIR = os.path.join(BUILD_DIR, "work")
    # copy files from filesdir into BUILD_DIR/files so that they are accessible
    # from within the sandbox
    FILESDIR = os.path.join(os.path.dirname(mod.FILE), "files")
    state.FILESDIR = os.path.join(BUILD_DIR, "files")
    if os.path.exists(FILESDIR):
        shutil.copytree(FILESDIR, state.FILESDIR)
    os.makedirs(state.WORKDIR, exist_ok=True)
    os.makedirs(state.T, exist_ok=True)

    state.ROOT = env.prefix().ROOT

    info(">>> " + l10n("pkg-unpacking"))
    # Network access is allowed exclusively during src_unpack, and
    # adds additional filesystem restrictions to the sandbox
    src_unpack(mod, BUILD_DIR, state.WORKDIR, state)

    default_basepath = mod.S or mod.get_default_source_basename()
    state.S = default_basepath or mod.P

    if default_basepath and os.path.exists(
        os.path.join(state.WORKDIR, default_basepath)
    ):
        WORKDIR = os.path.join(state.WORKDIR, default_basepath)
    else:
        WORKDIR = state.WORKDIR

    info(">>> " + l10n("pkg-preparing", dir=WORKDIR))

    src_prepare(mod, BUILD_DIR, WORKDIR, state)

    info(">>> " + l10n("pkg-prepared"))

    final_install = env.prefix().ROOT
    os.makedirs(final_install, exist_ok=True)

    state.D = os.path.join(BUILD_DIR, "image")
    os.makedirs(state.D, exist_ok=True)
    info(">>> " + l10n("pkg-installing-into", dir=state.D, atom=green(mod.ATOM.CPF)))
    src_install(mod, BUILD_DIR, WORKDIR, state)
    info(">>> " + l10n("pkg-installed-into", dir=state.D, atom=green(mod.ATOM.CPF)))

    os.chdir(env.TMP_DIR)

    if os.path.islink(state.D):
        installed_size = 0.0
    else:
        installed_size = get_tree_size(state.D) / 1024 / 1024

    build_size = get_tree_size(WORKDIR) / 1024 / 1024

    info("")
    info(f' {green("*")} ' + l10n("pkg-final-size-build", size=build_size))
    info(f' {green("*")} ' + l10n("pkg-final-size-installed", size=installed_size))
    info("")

    case_sensitive = not get_config_value("CASE_INSENSITIVE_FILES", False)
    old_pkg = load_installed_pkg(Atom(mod.CPN))

    info(">>> Checking for conflicts...")
    for (src, dst) in _iter_files_to_patch(
        state.D,
        final_install,
        case_sensitive=case_sensitive,
    ):
        if is_parent(dst, variable_data_dir()):
            raise PermissionError(l10n("local-dir-reserved", dir=variable_data_dir()))

        relative_path = os.path.normpath(os.path.relpath(dst, final_install))
        if (case_sensitive and os.path.exists(dst)) or (
            not case_sensitive and ci_exists(relative_path, prefix=env.prefix().ROOT)
        ):
            # Check if another package owns this file. If so, abort.
            conflict_pkg = belongs(relative_path, case_sensitive)
            if conflict_pkg and conflict_pkg != old_pkg:
                raise FileExistsError(
                    l10n(
                        "pkg-install-conflicting-file",
                        file=relative_path,
                        pkg=mod,
                        conflict_pkg=conflict_pkg,
                    )
                )

    # If a previous version of this mod was already installed,
    # remove it before doing the final copy
    db_path = os.path.join(vdb_path(), mod.CATEGORY, mod.PN)

    old_manifest = None

    if old_pkg:
        old_manifest = Manifest(os.path.join(db_path, "CONTENTS"))
        remove_pkg(old_pkg, os.path.exists(db_path) and mod.INSTALLED)

    info(
        ">>> "
        + l10n("pkg-installing-into", dir=final_install, atom=green(mod.ATOM.CPF))
    )

    contents_manifest = Manifest()

    overwrite_all = False

    for (src, dst) in _iter_files_to_patch(
        state.D,
        final_install,
        case_sensitive=case_sensitive,
    ):
        relative_path = os.path.normpath(os.path.relpath(dst, final_install))
        entry = ManifestEntry.from_path(FileType.MISC, src.path, relative_path)

        # Note that files which had not been modified since they were installed
        # will be removed when the old version is removed.
        # Files which have been modified won't, and thus will be protected here.
        if is_protected(dst) and os.path.exists(dst):
            # If the file is the same as in the previous version of the package
            # then ignore the file completely and leave the user's modified version
            if (
                old_manifest
                and old_manifest.get(relative_path, case_sensitive=case_sensitive)
                == entry
            ):
                continue

            # Otherwise, create a .new file and register it with the cfg_protect system
            new_dst = get_protected_path(dst)
            _add_redirection(new_dst, dst)
            dst = new_dst

        contents_manifest.add_entry(entry)
        if not os.path.isdir(os.path.dirname(dst)):
            os.makedirs(os.path.dirname(dst))
        try:
            _patch_file(src, dst, overwrite=False)
        except FileExistsError:
            # If not, prompt user to overwrite
            if env.INTERACTIVE and not overwrite_all:
                result = prompt_options(
                    l10n("pkg-file-conflict-prompt", file=relative_path),
                    [
                        (l10n("yes-short"), l10n("overwrite")),
                        (l10n("no-short"), l10n("dont-overwrite")),
                        (l10n("always-short"), l10n("always-overwrite")),
                    ],
                )
            else:
                result = l10n("yes-short")

            if result == l10n("always-short"):
                overwrite_all = True

            if result in (l10n("yes-short"), l10n("always-short")) or overwrite_all:
                preserved = dst + ".orig"
                info(l10n("pkg-file-conflict-overwrite", file=dst, preserved=preserved))
                os.rename(dst, preserved)
                _patch_file(src, dst, overwrite=True, move_function=shutil.copy2)

    state.ROOT = final_install
    pkg_postinst(mod, final_install, state.ROOT, state)

    # If installed database exists and there is no old mod, remove it
    if os.path.exists(db_path) and not old_pkg:
        shutil.rmtree(db_path, onerror=onerror)

    with VDB() as vdb:
        # Update db entry for installed mod
        os.makedirs(db_path, exist_ok=True)

        # Write CONTENTS
        contents_manifest.write(os.path.join(db_path, "CONTENTS"))
        vdb.git.add(os.path.join(db_path, "CONTENTS"))

        # Copy pybuild to DB
        # unless source pybuild is in DB (i.e we're reinstalling)
        if not mod.FILE.startswith(db_path):
            shutil.copy(mod.FILE, db_path)
            filespath = os.path.join(os.path.dirname(mod.FILE), "files")
            if os.path.exists(filespath):
                shutil.copytree(filespath, os.path.join(db_path, "files"))
                vdb.git.add(os.path.join(mod.CATEGORY, mod.PN, "files"))

        vdb.git.add(os.path.join(mod.CATEGORY, mod.PN, os.path.basename(mod.FILE)))

        manifest_path = os.path.join(os.path.dirname(mod.FILE), "Manifest")
        if os.path.exists(manifest_path):
            # Copy Manifest to DB
            if not mod.FILE.startswith(db_path):
                shutil.copy(manifest_path, db_path)
            vdb.git.add(os.path.join(mod.CATEGORY, mod.PN, "Manifest"))

        def add_installed(field: str, value: str):
            with open(os.path.join(db_path, field), "w") as use:
                print(value, file=use)
            vdb.git.add(os.path.join(mod.CATEGORY, mod.PN, field))

        # Copy installed use configuration to DB
        # Note: mod.get_use() may not be valid, as mod may be
        # an InstalledPybuild and we want to use the new configuration,
        # not the old one.
        add_installed("USE", " ".join(state.USE))
        # Copy repo pybuild was from to DB
        add_installed("REPO", mod.REPO)

        def fix_common(depstring: str):
            """
            Adds operator to dependencies in the common category
            to ensure this package is rebuilt if they change.
            """
            deps = depstring.split()
            for index, dep in enumerate(deps):
                if dep.startswith("common/"):
                    pkg = load_installed_pkg(Atom(dep))
                    # Note tilde operator. Revision bumps to
                    # common packages won't cause rebuilds.
                    if pkg:
                        deps[index] = f"~{dep}-{pkg.PV}"
            return " ".join(deps)

        # Store installed dependencies
        add_installed("DEPEND", fix_common(mod.DEPEND))

        # Copy pybuild environment to DB
        shutil.copy(
            os.path.join(_state_path(mod.FILE), "environment.xz"),
            os.path.join(db_path, "environment.xz"),
        )
        _delete_state(mod.FILE)
        vdb.git.add(os.path.join(mod.CATEGORY, mod.PN, "environment.xz"))

        if mod._PYBUILD_VER == 1:
            path = os.path.join(db_path, "environment.xz")
            if os.path.exists(path):
                compressed_environment = lzma.LZMAFile(path)
                try:
                    environment = json.load(compressed_environment)
                except EOFError as e:
                    raise RuntimeError(f"Failed to read {path}") from e

                if "REBUILD_FILES" in environment and environment["REBUILD_FILES"]:
                    manifest = Manifest()
                    for entry in get_rebuild_manifest(environment["REBUILD_FILES"]):
                        manifest.add_entry(entry)
                    manifest.write(os.path.join(db_path, "REBUILD_FILES"))
                    vdb.git.add(os.path.join(db_path, "REBUILD_FILES"))

        clear_cache_for_path(os.path.join(db_path, os.path.basename(mod.FILE)))

        os.chdir(old_curdir)
        print(">>> " + l10n("pkg-installed", atom=green(mod.ATOM.CPF)))
        info("")

        if not env.DEBUG:
            shutil.rmtree(BUILD_DIR, onerror=onerror)
            # Cleanup archive dir in case vfs had to extract anything
            _cleanup_tmp_archive_dir()
            info(">>> " + l10n("cleaned-up", dir=BUILD_DIR))

        clear_install_cache()
        return True



# Copyright 2022 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

from portmod.merge import global_updates
from portmodlib.l10n import l10n


def add_cfg_update_parser(subparsers, parents):
    parser = subparsers.add_parser(
        "cfg-update",
        help=l10n("cfg-update-help"),
        parents=parents,
        conflict_handler="resolve",
    )

    parser.set_defaults(func=lambda args: global_updates())



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
CLI prompt functions
"""

import re
import sys
from textwrap import fill
from typing import AbstractSet, List, Optional, Sequence, Tuple

from portmodlib.colour import bright, lblue, lgreen, lred
from portmodlib.l10n import l10n


def display_num_list(
    items: Sequence[str],
    notes: Optional[Sequence[str]] = None,
    selected: AbstractSet[int] = frozenset(),
    *,
    start: int = 0,
):
    """
    Displays a numbered list to stdout

    args:
        items: The primary elements to be displayed
        notes: A secondary note which, if provided, will be displayed in parentheses after the item
        selected: Selected indices will be followed by an asterisk
        start: The index to begin the list
    """
    padding = len(str(len(items)))

    def print_item(index: int, item: str, note: Optional[str]):
        selected_str = ""
        note_str = ""
        if index in selected:
            selected_str = lblue("*")
        index_str = bright("[" + str(index) + "]")
        if note:
            note_str = f" ({note})"
        print(
            f'  {index_str} {" " * (padding - len(str(index))) + item}',
            note_str,
            selected_str,
        )

    if notes:
        for index, (item, note) in enumerate(zip(items, notes)):
            print_item(index + start, item, note)
    else:
        for index, item in enumerate(items):
            print_item(index + start, item, None)


def strtobool(value: str) -> bool:
    """
    Returns true if the value is a string with contents representing true
    and false if the value is a string with contents representing false

    True options include: yes, y, true, t, 1
    False options include: no, n, false, f, 1

    This function is localized.
    """
    _true = {
        l10n("yes-short"),
        l10n("yes").lower(),
        l10n("true").lower(),
        l10n("true-short"),
        "1",
    }
    _false = {
        l10n("no-short"),
        l10n("no").lower(),
        l10n("false").lower(),
        l10n("false-short"),
        "0",
    }

    if value.lower() in _true:
        return True
    if value.lower() in _false:
        return False
    raise ValueError(f"Invalid boolean value {value}")


def prompt_bool(question) -> bool:
    """
    Prompts the user for yes or no

    See :func:`strtobool` for details of how the response is parsed

    args:
        question: The question to display prior to the prompt

    raises:
        EOFError: If the user enters -1 or EOF

    returns:
        True if the user answered yes, or false if the user answered no
    """

    sys.stdout.write(
        "{} [{}/{}]: ".format(
            question, bright(lgreen(l10n("yes"))), bright(lred(l10n("no")))
        )
    )
    while True:
        try:
            return strtobool(input().strip().lower())
        except ValueError:
            sys.stdout.write(
                l10n(
                    "prompt-invalid-response",
                    yes=bright(lgreen(l10n("yes"))),
                    no=bright(lred(l10n("no"))),
                )
            )


def prompt_options(question: str, options: List[Tuple[str, str]]) -> str:
    """
    Prompts the user for one of a given set of options

    args:
        question: The question to display prior to the prompt
        options: The list of possible option and description pairs

    raises:
        EOFError: If the user enters an EOF

    returns:
        The string entered by the user, which will correspond exactly to one of the options
    """
    print(question)
    # extra 2 for colon and an extra space
    max_option_width = max(len(option) for option, _ in options) + 2
    for option, desc in options:
        print(
            option + ":",
            fill(
                desc,
                width=80,
                initial_indent=" " * (max_option_width - len(option) - 2),
                subsequent_indent=" " * max_option_width,
            ),
        )
    sys.stdout.write("[{}]: ".format("/".join([option for option, _ in options])))
    option_set = {option for option, _ in options}
    while True:
        result = input().strip()
        if result in option_set:
            return result

        sys.stdout.write(
            l10n(
                "prompt-invalid-response-multiple",
                options="/".join([option for option, _ in options]),
            )
        )


def _parse_num_list(string: str, max_val: int) -> List[int]:
    if string == "":
        return []

    match = re.match(r"(\-?\d+)(?:-(\d+))?$", string)
    if not match:
        raise ValueError(l10n("prompt-invalid-range", max=max_val))
    start = match.group(1)
    end = match.group(2) or start
    return list(range(int(start, 10), int(end, 10) + 1))


def prompt_num_multi(question: str, max_val: int, cancel: bool = False) -> List[int]:
    """
    Prompts the user for one or more numbers

    Numbers must be positive, and must be separated by whitespace or commas.
    Ranges can be indicated using a hyphen

    args:
        question: The question to display prior to the prompt
        max_val: The maximum valid number that will be accepted
        cancel: If true, the user has the option to enter -1 to raise an EOFError

    raises:
        EOFError: If the user enters -1 (and cancel is True) or EOF

    returns:
        The list of numbers the user entered, with ranges resolved into the component integers
    """
    print(f"{question}: ")
    while True:
        try:
            result = [
                y
                for x in re.split("[ ,]+", input().strip())
                for y in _parse_num_list(x, max_val)
            ]
            if -1 in result and cancel:
                raise EOFError()
            if next(filter(lambda x: x > max_val or x < 0, result), None):
                print(l10n("prompt-range-too-large-number", max=max_val))
            else:
                return result
        except ValueError:
            print(l10n("prompt-invalid-range-multi", max=max_val))


def prompt_num(question: str, max_val: int, cancel: bool = False) -> int:
    """
    Prompts the user for exactly one number

    The input number must be positive

    args:
        question: The question to display prior to the prompt
        max_val: The maximum valid number that will be accepted
        cancel: If true, the user has the option to enter -1 to raise an EOFError

    raises:
        EOFError: If the user enters -1 (and cancel is True) or EOF

    returns:
        The number the user entered
    """

    print("{}: ".format(question))
    while True:
        try:
            result = int(input().strip())
            if result > max_val or result < 0:
                if result == -1 and cancel:
                    raise EOFError()

                print(l10n("prompt-range-too-large-number", max=max_val))
            else:
                return result
        except ValueError:
            print(l10n("prompt-invalid-range", max=max_val))



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
BLACKLIST and WHITELIST tests
"""

import os
import shutil
import zipfile

import pytest

from portmod.globals import env
from portmod.loader import load_file
from portmod.package import install_pkg
from portmod.repo import Repo
from portmodlib._deprecated import _get_install_dir_dest
from portmodlib.fs import get_hash

from .env import setup_env, tear_down_env
from .test_loader import TMP_REPO, create_pybuild

LISTED = set(
    map(
        os.path.normpath,  # type:  ignore
        ["directory/foo.txt", "directory/bar.txt", "directory/bar.png"],
    )
)
OTHER = set(map(os.path.normpath, ["foo.txt", "directory/bar.gif"]))  # type:  ignore


@pytest.fixture(autouse=True)
def setup():
    """
    Sets up and tears down the test environment
    """
    dictionary = setup_env("test")
    env.REPOS.append(Repo("test", TMP_REPO))
    yield dictionary
    tear_down_env()
    shutil.rmtree(TMP_REPO)


def create_zip():
    """Creates test zip file"""
    os.chdir(env.TMP_DIR)
    os.makedirs(env.DOWNLOAD_DIR, exist_ok=True)
    get_hash.cache_clear()
    if not os.path.exists(os.path.join(env.DOWNLOAD_DIR, "test.zip")):
        with zipfile.ZipFile(os.path.join(env.DOWNLOAD_DIR, "test.zip"), "w") as myzip:
            for file in LISTED | OTHER:
                if os.path.dirname(file):
                    os.makedirs(os.path.dirname(file), exist_ok=True)
                with open(file, "w") as filep:
                    print("", file=filep)
                myzip.write(file, file)  # type:  ignore


def test_whitelist(setup):
    """
    Tests that InstallDir whitelisting works properly
    """
    pybuild = """
import os
import sys
from pybuild import Pybuild1, InstallDir

class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"
    SRC_URI="test.zip"

    INSTALL_DIRS=[
        InstallDir(".", WHITELIST=["directory/*.txt", "directory/*.png", "*.jpg"]),
    ]
    """
    create_zip()
    file = create_pybuild(pybuild, manifest=True)
    mod = load_file(file)
    install_pkg(mod, set())

    base = os.path.join(env.prefix().ROOT, _get_install_dir_dest(mod))
    whitelisted = LISTED.copy()
    for root, _, files in os.walk(base):
        for file in files:
            path = os.path.relpath(root, base)
            fullpath = os.path.normpath(os.path.join(path, file))
            if fullpath in whitelisted:
                whitelisted.discard(fullpath)  # type:  ignore
            else:
                raise Exception(f"File {fullpath} is not in the whitelist!")

    # Only the whitelisted file can be present
    assert not whitelisted


def test_blacklist(setup):
    """
    Tests that InstallDir blacklisting works properly
    """
    pybuild = """
import os
import sys
from pybuild import Pybuild1, InstallDir

class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"
    SRC_URI="test.zip"

    INSTALL_DIRS=[
        InstallDir(".", BLACKLIST=["directory/*.txt", "directory/*.png", "*.jpg"]),
    ]
    """
    create_zip()
    file = create_pybuild(pybuild, manifest=True)
    mod = load_file(file)
    install_pkg(mod, set())

    other_files = OTHER.copy()

    base = os.path.join(env.prefix().ROOT, _get_install_dir_dest(mod))
    for root, _, files in os.walk(base):
        for file in files:
            path = os.path.relpath(root, base)
            fullpath = os.path.normpath(os.path.join(path, file))
            if fullpath in LISTED:
                raise Exception(f"File {fullpath} is in the black!")
            if fullpath in other_files:
                other_files.discard(fullpath)  # type:  ignore

    assert not other_files



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Tests the version comparison system
"""

import pytest

from portmodlib.atom import Version


def test_simple_version():
    """Tests extremely simple version comparison"""
    v_b = "2.0"
    v_a = "1.0"
    assert max([v_a, v_b]) == v_b


def gt(ver1, ver2):
    """Tests that ver1 is considered greater than ver2"""
    ver1 = Version(ver1)
    ver2 = Version(ver2)
    assert ver1 > ver2
    assert not ver2 > ver1
    assert max([ver1, ver2]) == ver1
    assert max([ver2, ver1]) == ver1


def test_suffix():
    """Tests that suffixed versions come before full versions"""
    assert max(
        map(Version, ["2.0_alpha", "2.0_beta", "2.0_pre", "2.0_rc", "2.0"])
    ) == Version("2.0")


def test_suffix_p():
    """Tests that p suffixed versions come after full versions"""
    gt("2.0_p1", "2.0")
    gt("2.0_alpha_p1", "2.0_alpha")


def test_suffix_endings():
    """Tests that p suffixed versions with integer endings order correctly"""
    gt("2.0_p2", "2.0_p1")
    gt("2.0_alpha2", "2.0_alpha1")
    gt("2.0_beta1", "2.0_alpha2")
    gt("2.0_alpha1", "2.0_alpha")
    gt("2.0_alpha1_beta2", "2.0_alpha")
    gt("2.0_alpha", "2.0_alpha_beta2")


def test_letter():
    """
    Tests that letter versions come after non-letter versions, increase in order
    and take precedence over suffixes, but not numeric components
    """
    gt("2.0a", "2.0")
    gt("2.0b", "2.0a")
    gt("2.0b_alpha", "2.0a")
    gt("2.1a", "2.0b")


def test_revision():
    """
    Tests that revisions have the lowest precedence and increase in order
    """
    gt("2.0-r1", "2.0")
    gt("2.0-r2", "2.0-r1")
    gt("2.0a-r1", "2.0-r2")
    gt("2.0_beta-r1", "2.0_alpha-r2")


def test_different_version_lengths():
    """
    Tests that having more version components is considered a greater version if the numbers are otherwise equal
    """
    gt("2.0.1", "2.0")
    gt("2.1", "2.0.1")
    gt("2.0.1_alpha", "2.0_p1-r1")


def test_different_epochs():
    """
    Tests that packages with a larger epoch are considered a greater version
    """
    gt("e1-2.0", "2.0")
    gt("e1-0.1", "2.0")
    gt("e2-0.1", "e1-2.0.0")
    gt("e1-2.1", "e1-2")


def test_leading_zeroes():
    """
    Tests that versions with leading zeroes are compared lexicographically
    """
    gt("2.1", "2.01")
    gt("2.10", "2.9")
    gt("2.1", "2.09")
    gt("2", "1.09")
    gt("2.09", "2")


def test_invalid():
    """
    Tests that invalid versions do not parse
    """
    with pytest.raises(TypeError):
        Version("1.")

    with pytest.raises(TypeError):
        Version("1.0-2")

    with pytest.raises(TypeError):
        Version("1.a")



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import subprocess
import sys
from pathlib import Path
from shutil import which
from typing import AbstractSet, Iterable, List, Set, Tuple, Union

from portmodlib.fs import is_parent
from portmodlib.l10n import l10n

from .config import get_config, get_config_value
from .globals import env
from .perms import Permissions
from .win32 import get_personal


def _required_exe(name: str) -> str:
    path = which(name)

    if not path:
        raise FileNotFoundError(l10n("missing-executable", file=name))
    return path


def resolve_paths(path: str) -> List[str]:
    """
    Returns all paths related to the symlink chain at paths
    If no symlinks exist, just returns path
    """
    paths = []
    prefix = ""
    for part in Path(path).parts:
        if os.path.islink(os.path.join(prefix, part)):
            last = os.path.join(prefix, part)
            paths.append(last)
            while os.path.islink(last):
                linkpath = os.readlink(last)
                if os.path.isabs(linkpath):
                    last = linkpath
                else:
                    last = os.path.normpath(
                        os.path.join(os.path.dirname(last), linkpath)
                    )
                paths.append(last)
        prefix = os.path.realpath(os.path.join(prefix, part))

    return paths


def parents(paths: Set[str]) -> Set[str]:
    results = set()
    for path in paths:
        for ppath in Path(path).parents:
            results.add(str(ppath))
    return results


def _dedup_subpaths(paths: Iterable[str]) -> List[str]:
    used: List[str] = []
    for path in sorted(paths):
        abspath = os.path.abspath(path)
        if os.path.exists(abspath):
            if any(is_parent(abspath, usedpath) for usedpath in used):
                continue
            used.append(abspath)

    return used


def get_paths(
    ro_paths: AbstractSet[str], rw_paths: AbstractSet[str]
) -> Tuple[Set[str], Set[str], Set[str]]:
    result_ro: Set[str] = set()
    result_rw: Set[str] = set()
    symlinks = set()

    def resolve_chain(path: str, path_set: Set[str]):
        chain = resolve_paths(path)
        if chain:
            # Elements are symlinks
            # the last is the end of the symlink chain.
            if os.path.islink(path):
                symlinks.add(path)

            path_set.add(os.path.realpath(path))
            for link in chain[:-1]:
                symlinks.add(link)
            # Ignore last element in chain, since it is
            # part of the resolved path (but not necessarily identical to the resolved
            # path if it's a directory symlink, but the original path pointed to either
            # a subdirectory of the symlink, or a file)
        else:
            path_set.add(path)

    for path in _dedup_subpaths(rw_paths):
        resolve_chain(path, result_rw)

    for path in _dedup_subpaths(ro_paths | rw_paths):
        if path not in result_rw:
            resolve_chain(path, result_ro)

    filtered_symlinks: Set[str] = set()
    for path in _dedup_subpaths(result_ro | result_rw | symlinks):
        if path in symlinks:
            filtered_symlinks.add(path)
    return result_ro, result_rw, filtered_symlinks


def sandbox_execute(
    command: List[str],
    permissions: Permissions,
    *,
    pipe_output: bool = False,
    err_on_stderr: bool = False,
    check: bool = True,
    pipe_error: bool = False,
) -> Union[str, None, bool]:
    """
    Executes command

    This is designed to be assigned to the FullPybuild.execute field prior to phase function
    execution. It is not set by default as it is not permitted outside of phase functions
    (e.g. in __init__)"""

    def cleanup():
        pass

    rw_paths = set(permissions.rw_paths)
    ro_paths = set(permissions.ro_paths)
    if env.TESTING:
        # Add source directory to whitelist so that we can get coverage from subprocesses
        rw_paths.add(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

    environ = os.environ.copy()

    # Set profile variables as environment variables in the sandbox
    for key, value in get_config().items():
        if isinstance(value, str):
            environ[key] = value
        elif isinstance(value, bool):
            if value:
                environ[key] = str(value)

    if permissions.tmp:
        environ["TMP"] = permissions.tmp
        os.makedirs(permissions.tmp, exist_ok=True)

    # ensure paths added at runtime are available
    if "PYTHONPATH" not in environ:
        environ["PYTHONPATH"] = ""
    if sys.platform == "win32":
        environ["PYTHONPATH"] += ";" + ";".join(sys.path)
    else:
        environ["PYTHONPATH"] += ":" + ":".join(sys.path)

    if sys.platform == "linux":
        if permissions.global_read:
            ro_paths.add("/")
        else:
            ro_paths |= {
                "/bin",
                "/etc",
                "/gnu",
                "/lib",
                "/lib32",
                "/lib64",
                "/run",
                "/nix",
                "/opt",
                "/usr",
                "/var",
            }

        ro_paths, rw_paths, symlinks = get_paths(ro_paths, rw_paths)

        ro_bind: List[str] = sum([["--ro-bind", path, path] for path in ro_paths], [])
        bind: List[str] = sum([["--bind", path, path] for path in rw_paths], [])
        symlink: List[str] = sum(
            [["--symlink", os.readlink(path), path] for path in symlinks], []
        )

        # Needs to be a list of command arguments
        # I.e. NOT a string
        user = get_config_value("SANDBOX_PERMISSIONS", [])
        if not isinstance(user, list):
            raise TypeError("_SANDBOX_PERMISSIONS must be a list of arguments to bwrap")

        sandbox_command = (
            [_required_exe("bwrap")]
            + ro_bind
            + bind
            + symlink
            + [
                "--dev",
                "/dev",
                "--proc",
                "/proc",
                "--unshare-all",
                "--chdir",
                os.getcwd(),
                "--die-with-parent",
            ]
            + user
        )
        if permissions.network:
            sandbox_command.append("--share-net")
        if permissions.tmp:
            sandbox_command.extend(["--bind", permissions.tmp, permissions.tmp])
        else:
            os.makedirs(env.PYBUILD_TMP_DIR, exist_ok=True)
            sandbox_command.extend(["--bind", env.PYBUILD_TMP_DIR, env.PYBUILD_TMP_DIR])
            environ["TMP"] = env.PYBUILD_TMP_DIR

    elif sys.platform == "win32":
        SINI = _required_exe("sbieini.exe")
        start_exe = _required_exe("start.exe")

        BOXNAME = "Portmod"
        delete_commands = []

        def win32_cleanup():
            olddir = os.getcwd()
            os.chdir(os.path.dirname(start_exe))
            try:
                subprocess.check_call(
                    f'"{start_exe}" /box:Portmod /silent /nosbiectrl delete_sandbox'
                )
            except subprocess.CalledProcessError:
                pass
            finally:
                for command in delete_commands:
                    subprocess.check_call(command)
                os.chdir(olddir)

        cleanup = win32_cleanup  # noqa

        def add_command(
            command: str, typ: str, value: str, delete: bool = True, boxname=BOXNAME
        ):
            nonlocal delete_commands
            subprocess.check_call(f'"{SINI}" {command} {boxname} {typ} "{value}"')
            if delete:
                delete_commands.append(f'"{SINI}" delete {boxname} {typ} "{value}"')

        if not permissions.network:
            add_command("set", "NetworkAccess", "*,Block")
            add_command("set", "AllowNetworkAccess", "n")
            add_command("set", "NotifyInternetAccessDenied", "n")
            # This enables a kernel-mode Windows Filtering Platform mechanism for
            # controlling network access, which should be more secure than the default
            # It requires reloading the driver, but we'll just enable it and hope the user
            # reloads at some point.
            add_command(
                "set", "NetworkEnableWFP", "y", delete=False, boxname="GlobalSettings"
            )

        # Create a temporary directory that can be used without affecting the user's system
        if not permissions.tmp:
            environ["TMP"] = env.PYBUILD_TMP_DIR
        add_command("append", "WriteFilePath", environ["TMP"])

        ro_paths.add(os.path.normpath(os.environ["programfiles(x86)"]))
        ro_paths.add(os.path.normpath(os.environ["programfiles"]))

        ro_paths, rw_paths, symlinks = get_paths(ro_paths, rw_paths)

        if not permissions.global_read:
            add_command("append", "ClosedFilePath", get_personal())

        for path in rw_paths:
            add_command("append", "OpenPipePath", path)

        # A wrapper command is used, combined with ForceProcess, as starting the process
        # using sandboxie's start.exe would not give us any stdout or stderr.
        add_command("set", "ForceFolder", os.getcwd())

        # Needs to be a list of command argument tuples
        user = get_config_value("SANDBOX_PERMISSIONS", [])
        if not isinstance(user, list):
            raise TypeError(
                "_SANDBOX_PERMISSIONS must be a list with sbieini commands in the order of $command $type $value. "
                'E.g. ["append", "ClosedFilePath", "filename.txt"]'
            )
        for user_command, typ, value in user:
            add_command(user_command, typ, value)
        add_command("set", "Enabled", "y", delete=False)

        sandbox_command = []
    elif sys.platform == "darwin":
        sandbox_command = [_required_exe("sandbox-exec"), "-p"]

        sandbox_profile = """
            (version 1)
            (deny default)
            (allow process-exec*)
            (allow signal (target self))
            (allow process-fork)
            (allow sysctl-read)
            (allow file-read*
                file-write-data
                (literal "/dev/null")
                (literal "/dev/zero")
            )
            (allow file-read*
                file-write-data
                file-ioctl
                (literal "/dev/dtracehelper")
            )
            (allow ipc-posix-shm)
        """

        if permissions.global_read:
            sandbox_profile += """
                (allow file-read*)
            """
        else:
            sandbox_profile += """
                (allow file-read-data file-read-metadata
                  (regex "^/dev/autofs.*")
                  (regex "^/Users/[^/]*/.gitconfig")
                  (literal "/var")
                  (literal "/dev/fd")
                  (literal "/dev/random")
                  (literal "/dev/urandom")
                  (literal "/var/db")
                  (literal "/var/db/xcode_select_link")
                  (literal "/")
                )
            """

            ro_paths |= {
                "/usr",
                "/System/Library",
                "/Applications",
                "/etc",
                "/private/etc",
                "/Library",
                "/opt",
            }

        if not permissions.tmp:
            environ["TMPDIR"] = env.PYBUILD_TMP_DIR
            os.makedirs(environ["TMPDIR"], exist_ok=True)
        rw_paths.add(environ["TMPDIR"])
        ro_paths.add("/private/var/select/sh")
        ro_paths, rw_paths, symlinks = get_paths(ro_paths, rw_paths)

        if not permissions.global_read:
            sandbox_profile += """
                (allow file-read*"""
            for path in ro_paths:
                sandbox_profile += f"""
                    (subpath "{path}")"""
            for path in parents(ro_paths | rw_paths):
                sandbox_profile += f"""
                    (literal "{path}")"""
            for path in symlinks:
                sandbox_profile += f"""
                    (literal "{path}")"""
            sandbox_profile += """
                )"""

        if permissions.network:
            sandbox_profile += "\n(allow network*)"

        sandbox_profile += """
            (allow file-write* file-read*"""
        for path in rw_paths:
            sandbox_profile += f"""
                (subpath "{path}")"""
        sandbox_profile += """
            )"""

        # Must be a string
        user = get_config_value("_SANDBOX_PERMISSIONS", "")
        if not isinstance(user, str):
            raise TypeError(
                "_SANDBOX_PERMISSIONS must be a string with sandbox-exec profile contents"
            )
        sandbox_profile += user

        sandbox_command.append(sandbox_profile)
    else:
        raise Exception("Unsupported Platform")

    output = None
    error = None
    if pipe_output:
        output = subprocess.PIPE
    if err_on_stderr or pipe_error:
        error = subprocess.PIPE

    try:
        proc = subprocess.run(
            sandbox_command + command,
            check=check,
            stdout=output,
            stderr=error,
            env=environ,
        )
    except subprocess.CalledProcessError as err:
        # Mostly arbitrary number used to indicate boolean true
        # Used by can-update-live
        if err.returncode == 142:
            return True
        raise err
    finally:
        cleanup()

    if err_on_stderr and proc.stderr:
        raise subprocess.CalledProcessError(
            0, sandbox_command + command, proc.stdout, proc.stderr
        )

    result = ""
    if pipe_output and proc.stdout:
        result += proc.stdout.decode("utf-8")
    if pipe_error and proc.stderr:
        result += proc.stderr.decode("utf-8")
    if pipe_output or pipe_error:
        return result

    # Mostly arbitrary number used to indicate boolean true
    if proc.returncode == 142:
        return True

    return None



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
from typing import Optional

from portmod.globals import env
from portmod.parsers.flags import add_flag, get_flags, remove_flag
from portmodlib.atom import QualifiedAtom

from ..config import get_config
from ..pybuild import Pybuild


def _user_package_accept_keywords_path() -> str:
    return os.path.join(env.prefix().CONFIG_DIR, "package.accept_keywords")


def add_keyword(
    atom: QualifiedAtom, keyword: str, *, protect_file: Optional[str] = None
):
    """Adds keyword for the given atom. Does not modify any existing keywords."""
    keyword_file = protect_file or _user_package_accept_keywords_path()

    add_flag(keyword_file, atom, keyword)


def remove_keyword(atom, keyword):
    keyword_file = os.path.join(env.prefix().CONFIG_DIR, "package.accept_keywords")
    remove_flag(keyword_file, atom, keyword)


def get_keywords(atom):
    keyword_file = os.path.join(env.prefix().CONFIG_DIR, "package.accept_keywords")
    ACCEPT_KEYWORDS = set(get_config()["ACCEPT_KEYWORDS"])
    return get_flags(keyword_file, atom).union(ACCEPT_KEYWORDS)


def accepts_stable(arch, keywords):
    return (
        # Package is always visible
        "**" in keywords
        # Package is visible if any architecture is stable
        or (
            "*" in keywords
            and any(x not in ("~*", "*") and x[0] not in ("-", "~") for x in keywords)
        )
        # Package is explicitly keyworded as stable for this architecture
        or any([keyword == arch for keyword in keywords])
    )


def accepts_testing(arch, keywords):
    return (
        accepts_stable(arch, keywords)
        # Package is explicitly keyworded as testing for this architecture
        or any(keyword == f"~{arch}" for keyword in keywords)
        # Package is visible if any architecture is testing
        or (
            "~*" in keywords
            and any(x not in ("~*", "*") and x[0] != "-" for x in keywords)
        )
    )


def accepts(accept_keywords, keywords):
    for keyword in accept_keywords:
        if keyword == "*":
            # Accepts stable on all architectures. Valid if keywords contains a stable
            # keyword for any keyword
            if any(keyword[0] not in ("~", "*", "-") for keyword in keywords):
                return True
        elif keyword == "~*":
            # Accepts testing on all architectures. Valid if keywords contains either
            # testing or stable for any keyword
            if any(not keyword.startswith("*") for keyword in keywords):
                return True
        elif keyword == "**":
            # Accepts any configuration
            return True
        elif keyword.startswith("~"):
            # Accepts testing on this architecture. Valid if keywords contains either
            # testing or stable for this keyword
            if accepts_testing(keyword[1:], keywords):
                return True
        else:  # regular keyword
            if accepts_stable(keyword, keywords):
                return True

    return "**" in keywords


def get_unstable_flag(mod: Pybuild) -> Optional[str]:
    """Returns the keyword for the user's current configuration of the given mod"""
    keywords = get_keywords(mod.ATOM)
    arch = env.prefix().ARCH
    if accepts(keywords, mod.KEYWORDS):
        return None

    if accepts_testing(arch, mod.KEYWORDS) or accepts_testing(arch, keywords):
        return "~"

    return "*"



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""Module for parsing and modifying sets such as @world"""

import os
from typing import Optional, Set

from portmod.config import variable_data_dir
from portmod.config.profiles import get_system
from portmod.globals import env
from portmodlib.atom import Atom, atom_sat

BUILTIN_SETS = {
    "world",
    "selected",
    "system",
    "selected-packages",
    "selected-sets",
    "rebuild",
    "installed",
}


def is_selected(atom: Atom) -> bool:
    """
    Returns true if and only if a package matching the given Atom is selected

    Selected packages are either system packages, packages included in the world file,
    packages included in the world sets file, or local packages
    """
    selected = get_set("world")
    for selatom in selected:
        if atom_sat(atom, selatom):
            return True
    return False


def get_set(package_set: str, parent_dir: Optional[str] = None) -> Set[Atom]:
    """
    Returns the atoms contained in the given package set

    Builtin sets include:

    * ``world``: The combination of the system and the selected sets
    * ``selected``: All packages explicitly selected, \
                either directly or indirectly through a set
    * ``system``: Packages required by the system profile
    * ``selected-packages``: Packages explicitly selected by the user
    * ``selected-sets``: Sets explicitly selected by the user
    * ``local-packages``: Packages installed locally by the user
    * ``rebuild``: Packages which have been selected for rebuilding \
               due to changes in the system
    * ``installed``: All installed packages

    args:
        package_set: The name of the set (note that this should not include the @
                     symbol used to reference sets on the command line
        parent_dir: The directory which stores the set file. If not specified
                    will be determined automatically. Non-builtin sets
                    are stored in ``env.prefix().SET_DIR``

    returns:
        The atoms contained in the given package set
    """
    if package_set == "world":
        return get_set("system") | get_set("selected")
    if package_set == "selected":
        selected = get_set("selected-packages") | get_set("local-packages")
        for selected_set in get_set("selected-sets"):
            selected |= get_set(selected_set)
        return selected
    if package_set == "system":
        return get_system()
    if package_set == "installed":
        from portmod.loader import load_all_installed

        return {pkg.ATOM for pkg in load_all_installed()}

    if parent_dir is None:
        parent_dir = env.prefix().SET_DIR

    if package_set == "local-packages":
        local_dir = env.prefix().LOCAL_MODS
        if os.path.exists(local_dir):
            return set(
                map(
                    lambda name: Atom(f"local/{name}"),
                    os.listdir(local_dir),
                )
            )
        return set()

    set_file = _get_set_path(package_set, parent_dir)
    if os.path.exists(set_file):
        with open(set_file, "r") as file:
            return {Atom(s) for s in file.read().splitlines()}
    return set()


def _get_set_path(package_set: str, parent_dir: Optional[str] = None) -> str:
    system_sets_dir = os.path.join(variable_data_dir(), "sets")
    if parent_dir is None:
        parent_dir = env.prefix().SET_DIR

    if package_set == "installed":
        raise RuntimeError("The @installed set cannot be modified by this function!")
    if package_set == "selected-packages":
        return os.path.join(system_sets_dir, "world")
    if package_set == "selected-sets":
        return os.path.join(system_sets_dir, "world_sets")
    if os.path.exists(os.path.join(system_sets_dir, package_set)):
        return os.path.join(system_sets_dir, package_set)

    return os.path.join(parent_dir, package_set)


def add_set(package_set: str, atom: Atom, parent_dir: Optional[str] = None):
    """
    Adds an atom to a set

    args:
        package_set: Name of the set to modify
        atom: Package atom to add to the set
        parent_dir: The directory where the set file is located.
                    If not specified, defaults to the prefix's set directory
                    (a subdirectory of the config directory).
    """
    set_file = _get_set_path(package_set, parent_dir)

    os.makedirs(os.path.dirname(set_file), exist_ok=True)
    if os.path.exists(set_file):
        with open(set_file, "r+") as file:
            for line in file:
                if atom in line:
                    break
            else:
                print(atom, file=file)
    else:
        with open(set_file, "a+") as file:
            print(atom, file=file)


def remove_set(package_set: str, atom: Atom, parent_dir: Optional[str] = None):
    """
    Removes an atom from a set

    args:
        package_set: Name of the set to modify
        atom: Package atom to remove from the set. Must match the expected atom exactly.
        parent_dir: The directory where the set file is located.
                    If not specified, defaults to the prefix's set directory
                    (a subdirectory of the config directory).
    """
    set_file = _get_set_path(package_set, parent_dir)
    if os.path.exists(set_file):
        with open(set_file, "r+") as file:
            new_f = file.readlines()
            file.seek(0)
            for line in new_f:
                if atom not in line:
                    file.write(line)
            file.truncate()



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Module for interacting with game configuration files as defined by
Config objects in the profile
"""

import csv
import os
from logging import warning
from typing import Dict, Set

from portmodlib.l10n import l10n


def read_userconfig(path: str) -> Dict[str, Set[str]]:
    """
    Parses csv-based user sorting rules

    args:
        path: Path of the file to be parsed
    returns:
        A dictionary mapping high-priority strings to strings they should override
    """
    userconfig = {}

    if os.path.exists(path):
        # Read user config
        with open(path, newline="") as csvfile:
            csvreader = csv.reader(csvfile, skipinitialspace=True)
            for row in csvreader:
                if row:
                    if len(row) == 1:
                        warning(l10n("user-config-warning", path=path, line=row[0]))
                    else:
                        atom = row[0].strip()
                        if atom not in userconfig:
                            userconfig[atom] = set(map(lambda x: x.strip(), row[1:]))
                        else:
                            userconfig[atom] |= set(map(lambda x: x.strip(), row[1:]))

    return userconfig



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import re
from dataclasses import dataclass
from typing import AbstractSet, Any, Dict, List, Match, Optional, Set, TypeVar

"""
Module for handling package atoms

All atom classes defined in this module should be considered read-only.
Modification of an Atom object may have unexpected side-effects
"""

flag_re = r"[A-Za-z0-9][A-Za-z0-9+_-]*"
useflag_re = re.compile(r"^" + flag_re + r"$")
usedep_re = (
    r"(?P<prefix>[!-]?)(?P<flag>"
    + flag_re
    + r")(?P<default>(\(\+\)|\(\-\))?)(?P<suffix>[?=]?)"
)
_usedep_re = re.compile("^" + usedep_re + "$")

op_re = r"(?P<B>(!!))?(?P<OP>([<>]=?|[<>=~]))?"
cat_re = r"((?P<C>[A-Za-z0-9][A-Za-z0-9\-]*)/)?"
ver_re = (
    r"(e(?P<EPOCH>\d+)-)?"
    r"(?P<NUMERIC>(\d+)(\.\d+)*)"
    r"(?P<LETTER>[a-z])?"
    r"(?P<SUFFIX>((_(pre|p|beta|alpha|rc)\d*)*))"
    r"(-r(?P<REV>\d+))?"
)
_ver_re = re.compile("^" + ver_re + "$")
repo_re = r"(::(?P<R>[A-Za-z0-9_][A-Za-z0-9_-]*(::installed)?))?"
_atom_re = re.compile(
    op_re
    + cat_re
    + r"(?P<P>(?P<PN>[A-Za-z0-9+_-]+?)(-(?P<PV>"
    + ver_re
    + r"))?)"
    + repo_re
    + r"(\[(?P<USE>.*)\])?$"
)


class InvalidAtom(Exception):
    "Exception indicating an atom has invalid syntax"


class UnqualifiedAtom(Exception):
    """
    Exception indicating an atom, which was expected to be
    qualified with a category, has no category
    """

    def __init__(self, atom):
        self.atom = atom

    def __str__(self):
        return f"Atom {self.atom} was expected to have a category!"


T = TypeVar("T", bound="Atom")


class Atom(str):
    @property
    def CP(self) -> Optional[str]:
        """
        The category, package name and version, excluding revision

        E.g. ``base/example-suite-1.0``
        """
        if self.CPN is None:
            return None
        if self.version is not None:
            return self.CPN + "-" + self.version.display(revision=True)
        return self.CPN

    CPN: Optional[str]
    """
    The category and package name

    E.g. ``base/example-suite``
    """
    USE: Set[str] = set()
    """
    Use flags set on the atom.

    E.g. ``{minimal}`` in ``base/example-suite[minimal]``
    """

    @property
    def P(self) -> str:
        """
        The package name and version, excluding revision

        E.g. ``example-suite-1.0``
        """
        return (
            self.PN
            + "-"
            + (self.version.display(revision=False) if self.version is not None else "")
        )

    PN: str
    """
    The package name

    E.g. ``example-suite``
    """

    @property
    def PF(self) -> str:
        """
        The package name and version, including revision

        E.g. ``example-suite-1.0-r1``
        """
        return self.PN + (
            "-" + self.version.display() if self.version is not None else ""
        )

    @property
    def PV(self) -> Optional[str]:
        """
        The package version, not including revision

        E.g. ``1.0``
        """
        if self.version is not None:
            return self.version.display(revision=False)
        return None

    @property
    def PR(self) -> Optional[str]:
        """
        The package revision

        E.g. ``r1``
        """
        if self.version is not None:
            return f"r{self.version.revision}"
        return None

    C: Optional[str]
    """
    The package category

    E.g. ``base``
    """
    R: Optional[str]
    """
    The package repository

    E.g. ``openmw`` in base/example-suite-1.0-r1::openmw
    """
    OP: Optional[str]
    """
    An operator.

    See PMS section `8.2.6.1 <https://projects.gentoo.org/pms/7/pms.html#x1-760008.2.6.1>`_,
    noting that Weak blocks are not currently supported by Portmod, and blocks don't show up
    in OP (they instead set the BLOCK field).

    E.g. ``openmw`` in base/example-suite-1.0-r1::openmw
    """
    BLOCK: bool
    """
    If true, this atom is a blocker, referring to a package which must not be installed.

    E.g. ``!!base/example-suite``
    """

    @property
    def PVR(self) -> Optional[str]:
        """
        The version and revision

        E.g. 1.0-r1
        """
        if self.version is not None:
            return self.version.display(revision=True)
        return None

    @property
    def CPF(self) -> str:
        """
        The category, package name and version, including revision

        E.g. ``base/example-suite-1.0-r1``
        """
        if self.CPN is None:
            prefix = self.PN
        else:
            prefix = self.CPN
        if self.version is not None:
            return prefix + "-" + self.version.display(revision=True)
        return prefix

    version: Optional["Version"] = None
    """
    The package's version
    """

    _CACHE: Dict[str, Any] = {}

    def __init__(self, atom: str):
        if atom in self._CACHE:
            self.__dict__ = self._CACHE[atom]
            return

        match = _atom_re.match(atom)
        if not match:
            raise InvalidAtom("Invalid atom %s. Cannot parse" % (atom))

        if match.group("PN") and match.group("C"):
            self.CPN = match.group("C") + "/" + match.group("PN")
        else:
            self.CPN = None

        if match.group("USE"):
            self.USE = set(match.group("USE").split(","))
            for x in self.USE:
                m = _usedep_re.match(x)
                if not m:
                    raise InvalidAtom(
                        "Invalid use dependency {} in atom {}".format(atom, x)
                    )

        if match.group("PV"):
            self.version = Version("", _match=match)

        self.PN = match.group("PN")
        self.C = match.group("C")
        self.R = match.group("R")
        self.OP = match.group("OP")
        self.BLOCK = match.group("B") is not None

        if self.OP is not None and self.version is None:
            raise InvalidAtom(
                "Atom %s has a comparison operator but no version!" % (atom)
            )

        self._CACHE[atom] = self.__dict__

    def evaluate_conditionals(self, use: AbstractSet[str]) -> "Atom":
        """
        Create an atom instance with any USE conditionals evaluated.
        @param use: The set of enabled USE flags
        @return: an atom instance with any USE conditionals evaluated
        """
        tokens = set()

        for x in self.USE:
            m = _usedep_re.match(x)

            if m is not None:
                operator = m.group("prefix") + m.group("suffix")
                flag = m.group("flag")
                default = m.group("default")
                if default is None:
                    default = ""

                if operator == "?":
                    if flag in use:
                        tokens.add(flag + default)
                elif operator == "=":
                    if flag in use:
                        tokens.add(flag + default)
                    else:
                        tokens.add("-" + flag + default)
                elif operator == "!=":
                    if flag in use:
                        tokens.add("-" + flag + default)
                    else:
                        tokens.add(flag + default)
                elif operator == "!?":
                    if flag not in use:
                        tokens.add("-" + flag + default)
                else:
                    tokens.add(x)
            else:
                raise Exception("Internal Error when processing atom conditionals")

        atom = Atom(self)
        atom.USE = tokens
        return atom

    def strip_use(self: T) -> T:
        """Returns the equivalent of this atom with the USE dependencies removed"""
        return self.__class__(re.sub(r"\[.*\]", "", str(self)))

    def use(self, *flags: str):
        """returns atom with use flag dependency"""
        return self.__class__(f'{self}[{",".join(flags)}]')

    def with_category(self, category: str) -> "QualifiedAtom":
        return QualifiedAtom(
            (self.BLOCK and "!!" or "")
            + (self.OP or "")
            + category
            + "/"
            + self.PF
            + (self.R or "")
        )


class QualifiedAtom(Atom):
    """Atoms that include categories"""

    CP: str
    CPN: str
    CPF: str
    C: str

    def __init__(self, atom: str):
        super().__init__(atom)

        if not self.C:
            raise UnqualifiedAtom(atom)


class VAtom(QualifiedAtom):
    """Atoms that include version information"""

    PV: str
    PVR: str
    version: "Version"

    def __init__(self, atom: str):
        super().__init__(atom)
        assert self.version


class FQAtom(VAtom):
    """Atoms that include all possible non-optional fields"""

    R: str

    def __init__(self, atom: str):
        super().__init__(atom)
        assert self.R


@dataclass
class Version:
    """
    A package version

    This class should be treated as immutable
    """

    numeric: List[str]
    suffixes: List[str]
    epoch: Optional[int] = None
    letter: str = ""
    revision: Optional[int] = None

    def __init__(self, version: str, *, _match: Optional[Match[str]] = None):
        """
        args:
            version: A string representation of the version to be parsed
        """
        if not _match:
            match = _ver_re.match(version)
            if match is None:
                raise TypeError(f"Version {version} could not be parsed")
        else:
            match = _match

        if match.group("EPOCH"):
            self.epoch = int(match.group("EPOCH"))
        self.numeric = list(str(match.group("NUMERIC")).split("."))
        self.letter = match.group("LETTER") or ""
        suffixes = match.group("SUFFIX")
        if suffixes:
            self.suffixes = suffixes.lstrip("_").split("_")
        else:
            self.suffixes = []
        if match.group("REV"):
            self.revision = int(match.group("REV"))

    def __str__(self) -> str:
        return self.display()

    def display(self, *, revision: bool = True, epoch: bool = True) -> str:
        return (
            (f"e{self.epoch}-" if self.epoch is not None and epoch else "")
            + ".".join(self.numeric)
            + (self.letter or "")
            + ("_" + "_".join(self.suffixes) if self.suffixes else "")
            + (f"-r{self.revision}" if self.revision is not None and revision else "")
        )

    def _normalize_numeric(self) -> List[str]:
        return list(map(lambda x: x.rstrip("0"), self.numeric))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Version):
            return self.compare(other)
        return False

    def compare(self, other: "Version", *, ignore_revision=False) -> bool:
        # Note: versions with different numbers of numeric components are not equivalent
        # according to the PMS, even if the extra components are trailing 0s.
        return (
            self.epoch == other.epoch
            and self._normalize_numeric() == other._normalize_numeric()
            and self.letter == other.letter
            and self.suffixes == other.suffixes
            and (self.revision == other.revision or ignore_revision)
        )

    def __gt__(self, other: "Version") -> bool:
        # Compare epochs
        if (self.epoch or 0) > (other.epoch or 0):
            return True
        if (self.epoch or 0) < (other.epoch or 0):
            return False

        if int(self.numeric[0]) > int(other.numeric[0]):
            return True
        if int(self.numeric[0]) < int(other.numeric[0]):
            return False

        # Compare numeric components
        for index, val in enumerate(self.numeric[1:], start=1):
            if index >= len(other.numeric):
                return True
            if val.startswith("0") or other.numeric[index].startswith("0"):
                # If either starts with a leading 0, strip trailing zeroes and
                # compare lexicographically.
                # I.e. 01 should always come before 10
                my_stripped = val.rstrip("0")
                other_stripped = other.numeric[index].rstrip("0")
                if my_stripped > other_stripped:
                    return True
                if my_stripped < other_stripped:
                    return False
            else:
                if int(val) > int(other.numeric[index]):
                    return True
                if int(val) < int(other.numeric[index]):
                    return False
        if len(other.numeric) > len(self.numeric):
            return False

        # Compare letter components
        if self.letter > other.letter:
            return True
        if self.letter < other.letter:
            return False

        # Compare suffixes
        for a_s, b_s in zip(self.suffixes, other.suffixes):
            asm = re.match(r"(?P<S>[a-z]+)(?P<N>\d+)?", a_s)
            bsm = re.match(r"(?P<S>[a-z]+)(?P<N>\d+)?", b_s)
            assert asm
            assert bsm
            a_suffix = asm.group("S")
            b_suffix = bsm.group("S")
            a_suffix_num = int(asm.group("N") or "0")
            b_suffix_num = int(bsm.group("N") or "0")
            if a_suffix == b_suffix:
                if b_suffix_num > a_suffix_num:
                    return False
                if a_suffix_num > b_suffix_num:
                    return True
            elif suffix_gt(a_suffix, b_suffix):
                return True
            else:
                return False
        # More suffixes implies an earlier version,
        # except when the suffix is _p
        if len(self.suffixes) > len(other.suffixes):
            if self.suffixes[len(other.suffixes)].startswith("p"):
                return True
            return False
        if len(self.suffixes) < len(other.suffixes):
            if other.suffixes[len(self.suffixes)].startswith("p"):
                return False
            return True

        # Compare revisions
        if (self.revision or 0) > (other.revision or 0):
            return True
        if (self.revision or 0) < (other.revision or 0):
            return False

        # Equal
        return False


def suffix_gt(a_suffix: str, b_suffix: str) -> bool:
    """Returns true iff a_suffix > b_suffix"""
    suffixes = ["alpha", "beta", "pre", "rc", "p"]
    return suffixes.index(a_suffix) > suffixes.index(b_suffix)


def version_gt(version1: str, version2: str) -> bool:
    """
    Version comparision function

    args:
        version1: A version string
        version2: Another version string

    returns:
        True if and only if version1 is greater than version2
    """
    return Version(version1) > Version(version2)


def atom_sat(specific: Atom, generic: Atom, *, ignore_name: bool = False) -> bool:
    """
    Determines if a fully qualified atom (can only refer to a single package)
    satisfies a generic atom
    """

    if not ignore_name:
        if specific.PN != generic.PN:
            # Mods must have the same name
            return False

        if generic.C and (generic.C != specific.C):
            # If para defines category, it must match
            return False

    if generic.R and (generic.R != specific.R):
        # If para defines repo, it must match
        return False

    if not generic.OP:
        # Simple atom, either one version or all versions will satisfy

        # Check if version is correct
        if generic.version and (
            not specific.version
            or not specific.version.compare(generic.version, ignore_revision=True)
        ):
            return False

        # Check if revision is correct
        if (
            generic.version
            and generic.version.revision
            and (
                not specific.version
                or specific.version.revision != generic.version.revision
            )
        ):
            return False
    elif generic.version and specific.version:
        equal = specific.version == generic.version
        verequal = specific.version.compare(generic.version, ignore_revision=True)
        lessthan = specific.version < generic.version
        greaterthan = specific.version > generic.version

        if generic.OP == "=":
            return equal
        if generic.OP == "~":
            return verequal
        if generic.OP == "<":
            return lessthan
        if generic.OP == "<=":
            return equal or lessthan
        if generic.OP == ">":
            return greaterthan
        if generic.OP == ">=":
            return equal or greaterthan

    return True



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
from logging import error, warning

from portmod.globals import env
from portmod.loader import load_installed_pkg
from portmod.parsers.manifest import ManifestEntry
from portmod.vdb import vdb_path
from portmodlib.atom import Atom
from portmodlib.l10n import l10n


def get_packages(path: str):
    for category in os.listdir(path):
        if os.path.isdir(os.path.join(path, category)) and not category.startswith("."):
            for package in os.listdir(os.path.join(path, category)):
                if os.path.isdir(os.path.join(path, category, package)):
                    yield category, package


def validate(args):
    # Check that mods in the DB correspond to mods in the mods directory
    for category, package in get_packages(vdb_path()):
        # Check that pybuild can be loaded
        pkg = load_installed_pkg(Atom(f"{category}/{package}"))
        if not pkg:
            error(
                l10n(
                    "in-database-could-not-load",
                    atom=Atom(f"{category}/{package}"),
                )
            )
        else:
            # Check files listed in CONTENTS
            for path, entry in pkg.get_contents().entries.items():
                realpath = os.path.join(env.prefix().ROOT, path)
                if not os.path.exists(realpath):
                    warning(l10n("installed-file-missing", path=path))
                elif entry != ManifestEntry.from_path(entry.filetype, realpath, path):
                    warning(l10n("installed-file-mismatch", path=path, pkg=pkg.CPN))



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

from typing import Any, Dict, List, Tuple, Union, cast


class CycleException(Exception):
    """Exception indicating a graph cycle"""

    def __init__(self, message, cycle: Union[List[str], List[Tuple[str]]]):
        if isinstance(cycle[0], tuple):
            cyclestr = " -> ".join(map(lambda x: x[0], cycle))
        else:
            cyclestr = " -> ".join(cast(List[str], cycle))
        super().__init__(message + " " + cyclestr)
        self.cycle = cycle


def __min_priority(priorities, nodes):
    priority = "z"
    resultset = set()
    for node in nodes:
        if priorities[node] < priority:
            priority = priorities[node]
            resultset = set([node])
        elif priorities[node] == priority:
            resultset.add(node)

    return resultset


# Topological sorting based on Kahn's algorithm, with the smallest elements
# appearing early to ensure consisteny between runs
# Graph is a dictionary mapping each vertex to a set containing its children.
# Priority is a dictionary mapping each vertex to a priority in [0-9z]
def tsort(graph, priority=None):
    _graph = graph.copy()
    L = []
    S = {node for node in _graph.keys()}

    for edges in _graph.values():
        S -= edges

    while len(S) > 0:
        # We always take the smallest value from the set, rather than an arbitrary value
        if priority is None:
            smallest = min(S)
        else:
            smallest = min(__min_priority(priority, S))

        S.remove(smallest)
        L.append(smallest)

        s_set = _graph[smallest]
        _graph[smallest] = set()
        for node in s_set:
            if not any([node in edges for edges in _graph.values()]):
                S.add(node)

    if any([edges for edges in _graph.values()]):
        # Graph has at least one cycle

        def invert(g):
            new: Dict[Any, Any] = {}
            for node in g:
                for link in g[node]:
                    if link in new:
                        new[link].add(node)
                    else:
                        new[link] = {node}
            return new

        def search(g, node, previous):
            if node in previous:
                i = previous.index(node)
                return previous[i:] + [node]

            for other in g[node]:
                result = search(g, other, previous + [node])
                if result:
                    return result
            return None

        # Invert graph and search for cycle
        cycle = search(
            invert(_graph), next(node for node in _graph if _graph[node]), []
        )
        raise CycleException("There is a cycle in the graph!", cycle)

    return L



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import re
import sys
from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    cast,
)

from portmodlib.atom import Atom, FQAtom, QualifiedAtom, atom_sat
from portmodlib.colour import blue, bright, green, lgreen, red, yellow
from portmodlib.l10n import l10n
from portmodlib.usestr import use_reduce

from .config import get_config
from .config.sets import is_selected
from .config.use import get_use
from .download import fetchable, find_download, get_download_size
from .loader import (
    SandboxedError,
    _sandbox_execute_pybuild,
    load_installed_pkg,
    load_pkg_fq,
)
from .parsers.flags import collapse_flags
from .perms import Permissions
from .pybuild import InstalledPybuild, Pybuild
from .query import get_flag_string
from .tsort import CycleException, tsort
from .util import select_package


class PackageDoesNotExist(Exception):
    """Indicates that no mod matching this atom could be loaded"""

    def __init__(self, atom: Optional[Atom] = None, *, msg=None):
        super().__init__(msg or l10n("package-does-not-exist", atom=green(atom)))


class Trans:
    """Transaction class"""

    REPR: str
    COLOUR: Callable[[str], str]
    pkg: Pybuild
    flags: Set[str]

    def __init__(self, pkg: Pybuild, flags: Iterable[str]):
        self.pkg = pkg
        self.flags = set(flags)

    def __str__(self):
        return f"{self.__class__.__name__}({self.pkg})"

    def __repr__(self):
        return str(self)


class Delete(Trans):
    """Delete Transaction"""

    REPR = "d"
    COLOUR = red
    pkg: InstalledPybuild

    def __init__(self, pkg: InstalledPybuild):
        super().__init__(pkg, set())


class New(Trans):
    """New Package Transaction"""

    REPR = "N"
    COLOUR = lgreen


class Change(Trans):
    """Update Package Transaction"""

    COLOUR = blue

    def __init__(self, pkg: Pybuild, old: InstalledPybuild, flags: Iterable[str]):
        super().__init__(pkg, flags)
        self.old = old


class Update(Change):
    """Downgrade Package Transaction"""

    REPR = "U"


class Downgrade(Change):
    """Downgrade Package Transaction"""

    REPR = "D"


class Reinstall(Trans):
    """Reinstall Package Transaction"""

    REPR = "R"
    COLOUR = yellow


class Transactions:
    pkgs: List[Trans]
    config: Set[Any]
    new_selected: Set[Pybuild]

    def __init__(self):
        self.pkgs = []
        self.config = set()
        self.new_selected = set()

    def copy(self) -> "Transactions":
        new = Transactions()
        new.pkgs = self.pkgs.copy()
        new.config = self.config.copy()
        new.new_selected = self.new_selected.copy()
        return new

    def append(self, trans: Trans):
        self.pkgs.append(trans)

    def add_new_selected(self, pkg: Pybuild):
        self.new_selected.add(pkg)

    def extend(self, trans: "Transactions"):
        self.pkgs.extend(trans.pkgs)
        self.config |= trans.config
        self.new_selected |= trans.new_selected


class UseDep:
    def __init__(
        self, atom: FQAtom, flag: str, oldvalue: Optional[str], comment: Sequence[str]
    ):
        self.atom = atom
        self.flag = flag
        self.oldvalue = oldvalue
        self.comment = comment

    def __repr__(self):
        if self.oldvalue:
            return f"UseDep({self.atom}, {self.oldvalue} -> {self.flag})"
        else:
            return f"UseDep({self.atom}, {self.flag})"


def get_usestrings(
    mod: Pybuild,
    installed_use: Optional[Set[str]],
    enabled_use: Set[str],
    verbose: bool,
) -> List[str]:
    # Note: flags containing underscores are USE_EXPAND flags
    # and are displayed separately
    IUSE_STRIP = {flag.lstrip("+") for flag in mod.IUSE if "_" not in flag}

    texture_options = use_reduce(
        mod.TEXTURE_SIZES, enabled_use, flat=True, token_class=int
    )

    use_expand_strings = []
    for use in get_config().get("USE_EXPAND", []):
        if use in get_config().get("USE_EXPAND_HIDDEN", []):
            continue
        prefix = use.lower() + "_"

        def base_flag(flag: str):
            return re.sub(f"^{prefix}", "", flag)

        enabled_expand = {
            base_flag(flag) for flag in enabled_use if flag.startswith(prefix)
        }
        disabled_expand = {
            base_flag(flag)
            for flag in mod.IUSE_EFFECTIVE
            if flag.startswith(prefix) and base_flag(flag) not in enabled_expand
        }
        if enabled_expand or disabled_expand:
            installed_expand: Optional[Set[str]]
            if installed_use is not None:
                installed_expand = {
                    base_flag(flag) for flag in installed_use if flag.startswith(prefix)
                }
            else:
                installed_expand = None
            string = get_flag_string(
                use, enabled_expand, disabled_expand, installed_expand, verbose=verbose
            )
            use_expand_strings.append(string)

    if mod.TEXTURE_SIZES is not None and len(texture_options) >= 2:
        texture_size = next(
            (
                use.lstrip("texture_size_")
                for use in enabled_use
                if use.startswith("texture_size")
            ),
            None,
        )
        if texture_size is not None:
            texture_string = get_flag_string(
                "TEXTURE_SIZE",
                [texture_size],
                map(str, sorted(set(texture_options) - {int(texture_size)})),
            )
        else:
            texture_string = ""
    else:
        texture_string = ""

    usestring = get_flag_string(
        "USE",
        enabled_use & IUSE_STRIP,
        IUSE_STRIP - enabled_use,
        installed_use,
        verbose=verbose,
    )

    return [usestring] + use_expand_strings + [texture_string]


def print_transactions(
    transactions: Transactions,
    verbose: bool = False,
    out=sys.stdout,
    summarize: bool = True,
):
    pkgs = transactions.pkgs

    for trans in pkgs:
        pkg = trans.pkg
        installed_mod = load_installed_pkg(Atom(trans.pkg.CPN))
        if installed_mod is None:
            installed_use = None
        else:
            installed_use = installed_mod.INSTALLED_USE

        v = verbose or isinstance(trans, New)

        if isinstance(trans, Delete):
            usestring = ""
        else:
            usestrings = get_usestrings(pkg, installed_use, trans.flags, v)
            usestring = " ".join(list(filter(None, usestrings)))

        trans_colour = trans.__class__.COLOUR
        oldver = ""
        if isinstance(trans, Change):
            oldver = blue(" [" + trans.old.PVR + "]")

        modstring: str
        if verbose:
            modstring = pkg.ATOM
        else:
            modstring = pkg.ATOM.CPF

        if is_selected(pkg.ATOM) or pkg in transactions.new_selected:
            modstring = bright(green(modstring))
        else:
            modstring = green(modstring)

        fetch_status = " "
        fetchable_files = set(fetchable(trans.pkg, trans.flags))
        fetch_restricted = [
            source
            for source in trans.pkg.get_source_manifests(trans.flags)
            if source not in fetchable_files
        ]
        if fetch_restricted:
            if all(find_download(source) for source in fetch_restricted):
                fetch_status = "f"
            else:
                fetch_status = "F"

        print(
            f"[{red(fetch_status)}{bright(trans_colour(trans.REPR))}]"
            f" {modstring}{oldver} {usestring}",
            file=out,
        )

    if summarize:
        download_size = get_download_size(
            {trans.pkg: trans.flags for trans in pkgs if not isinstance(trans, Delete)}
        )
        print(
            l10n(
                "transaction-summary",
                packages=len(pkgs),
                updates=len([trans for trans in pkgs if isinstance(trans, Change)]),
                new=len([trans for trans in pkgs if isinstance(trans, New)]),
                reinstalls=len(
                    [trans for trans in pkgs if isinstance(trans, Reinstall)]
                ),
                removals=len([trans for trans in pkgs if isinstance(trans, Delete)]),
                download=download_size,
            ),
            file=out,
        )


def get_all_deps(depstring: str) -> List[Atom]:
    dependencies = use_reduce(depstring, token_class=Atom, matchall=True, flat=True)

    # Note that any || operators will still be included. strip those out
    return list(
        [dep for dep in dependencies if dep != "||" and not dep.startswith("!")]
    )


def sort_transactions(transactions: Transactions):
    """
    Create graph and do a topological sort to ensure that mods are installed/removed
    in the correct order given their dependencies
    """

    def get_dep_graph(rdepend=True):
        graph: Dict[Atom, Set[Atom]] = {}
        priorities = {}

        for trans in transactions.pkgs:
            graph[trans.pkg.ATOM] = set()
            # Always remove packages last
            # FIXME: This is a poor workaround to the fact that the removal order
            # doesn't ensure that the package being replaced has RDEPEND satisfied
            # something which is necessary for pkg_prerm (note that it may not be
            # possible if the old and new versions have dependencies which cannot
            # be installed simultaneously)
            if isinstance(trans, Delete):
                priorities[trans.pkg.ATOM] = "z"
            else:
                if trans.pkg._PYBUILD_VER == 1:
                    priorities[trans.pkg.ATOM] = trans.pkg.TIER  # type: ignore
                else:
                    priorities[trans.pkg.ATOM] = "a"

        def add_depends(mod, key: str, delete: bool):
            depends = {}
            depstring = getattr(mod, key)
            for dep in get_all_deps(depstring):
                for trans in transactions.pkgs:
                    if atom_sat(trans.pkg.ATOM, dep):
                        depends[trans.pkg.ATOM] = trans.pkg

            if delete:
                # When removing packages, remove them before their dependencies
                graph[mod.ATOM] |= set(depends.keys())
            else:
                # When adding or updating packages, packages, add or update their dependencies
                # before them
                for dep in depends:
                    graph[dep].add(mod.ATOM)
                    if key == "DEPEND":
                        # Also ensure runtime dependencies are available for build dependencies
                        # Whether or not we enforce runtime dependencies for all packages
                        add_depends(depends[dep], "RDEPEND", False)

        for trans in transactions.pkgs:
            add_depends(trans.pkg, "DEPEND", isinstance(trans, Delete))
            if rdepend:
                add_depends(trans.pkg, "RDEPEND", isinstance(trans, Delete))
        return graph, priorities

    # Attempt to sort using both runtime and build dependencies. If this fails,
    # fall back to just build dependencies
    graph, priorities = get_dep_graph()
    try:
        mergeorder = tsort(graph, priorities)
    except CycleException:
        try:
            graph, priorities = get_dep_graph(rdepend=False)
            mergeorder = tsort(graph, priorities)
        except CycleException as exception:
            raise CycleException(
                l10n("cycle-encountered-when-sorting-transactions"), exception.cycle
            )

    new_trans = transactions.copy()
    new_trans.pkgs = []
    for atom in mergeorder:
        for trans in transactions.pkgs:
            if trans.pkg.ATOM == atom:
                new_trans.pkgs.append(trans)
                break

    return new_trans


def generate_transactions(
    enabled: Iterable[FQAtom],
    disabled: Iterable[FQAtom],
    newselected: AbstractSet[QualifiedAtom],
    usedeps: Iterable[UseDep],
    enabled_flags: Dict[FQAtom, Set[str]],
    *,
    emptytree: bool = False,
    update: bool = False,
) -> Transactions:
    """
    Generates a list of transactions to update the system such that
    all packages in enabled are installed and all packages in disabled are not

    Packages will not be rebuilt unless a change has occurred, or they are included
    in the new_selected parameter set and update is not specified.

    args:
        enabled: Packages that should be enabled, if not already
        disabled: Packages that should be disabled, if not already
        new_selected: Packages that were selected by the user for this operation
                      These should be re-installed, even if no change has been
                      made, unless update is also passed
        usedeps: Use changes that should accompany the transactions
        update: If true, will update live packages, and won't re-install packages which haven't changed

    returns:
        Onject representing the transactions
    """

    transactions = Transactions()

    # Note: while this technically requires knowing about the use flags in advance,
    # despite being needed to determine the use flags enabled, any flags which are related
    # to a relevant alias would have been included in flagupdates as they are handled by
    # the dependency calculator.
    def is_installed(atom: Atom) -> bool:
        for fqatom in enabled:
            if atom_sat(fqatom, atom) and atom.USE <= enabled_flags[fqatom]:
                return True
        return False

    for atom in enabled:
        pkg = load_pkg_fq(atom)

        flags = set(
            filter(
                lambda x: not x.startswith("-"),
                collapse_flags(
                    get_use(pkg, is_installed=is_installed)[0], enabled_flags[atom]
                ),
            )
        )

        if "local" in pkg.PROPERTIES:
            continue

        (to_install, dep) = select_package([pkg])

        if dep is not None:
            if isinstance(dep, (list, set)):
                transactions.config |= set(dep)
            else:
                transactions.config.add(dep)

        installed = load_installed_pkg(Atom(atom.CPN))

        if not (to_install or installed):
            raise PackageDoesNotExist(atom)

        if (
            to_install is not None
            and to_install.ATOM.CPN in newselected
            or (installed and installed.ATOM.CPN in newselected)
        ):
            transactions.add_new_selected(cast(Pybuild, to_install or installed))

        if emptytree:
            transactions.append(
                Reinstall(cast(Pybuild, to_install or installed), flags)
            )
            continue

        # TODO: There might be advantages to preferring installed over to_install
        # such as avoiding re-downloading files just because the sources changed in a trivial
        # manner

        def can_update_live(pkg: InstalledPybuild):
            try:
                return _sandbox_execute_pybuild(
                    pkg.FILE,
                    "can-update-live",
                    Permissions(network=True),
                    installed=True,
                )
            except SandboxedError:
                return False

        if installed is not None:
            if installed.version > (to_install or installed).version:
                # to_install cannot be None if it has a smaller version
                transactions.append(Downgrade(to_install, installed, flags))
            elif ((to_install or installed).version > installed.version) or (
                update
                and "live"
                in use_reduce(installed.PROPERTIES, installed.INSTALLED_USE, flat=True)
                and can_update_live(installed)
            ):
                transactions.append(Update(to_install or installed, installed, flags))
            elif installed.INSTALLED_USE != flags:
                transactions.append(Reinstall(to_install or installed, flags))
            elif not update and installed.ATOM.CPN in newselected:
                transactions.append(Reinstall(to_install or installed, flags))
            elif to_install and not atom.R.endswith("::installed"):
                # If the repo's version is enabled, this means the dep generator
                # wants the package re-installed using the repo's version for some reason
                transactions.append(Reinstall(to_install, flags))
        elif to_install is not None:
            new_mod = to_install
            transactions.append(New(new_mod, flags))

    for atom in disabled:
        to_remove = load_installed_pkg(Atom(atom))
        if to_remove is not None:
            transactions.append(Delete(to_remove))

    for dep in usedeps:
        for trans in transactions.pkgs:
            if atom_sat(trans.pkg.ATOM, dep.atom):
                transactions.config.add(dep)

    return transactions



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
from typing import Dict

from portmod.repo import Repo, get_repo
from portmod.repo.metadata import get_master_names
from portmod.repos import add_repo, get_local_repos, get_remote_repos
from portmod.sync import sync
from portmodlib.l10n import l10n


def sync_args(args):
    local_repos = get_local_repos()
    remote_repos = get_remote_repos()
    to_sync: Dict[str, Repo] = {}

    meta = get_repo("meta")
    # Always sync meta, in case it needs to inform us
    # of newly added repositories
    if not os.path.exists(meta.location):
        sync([meta])
    else:
        to_sync["meta"] = meta

    def add_repo_and_deps(repo_name):
        if repo_name in to_sync:
            return

        repo = local_repos.get(repo_name) or add_repo(remote_repos[repo_name])

        if repo:
            for dep in get_master_names(repo.location):
                add_repo_and_deps(dep)
            to_sync[repo_name] = repo
        else:
            to_sync[repo_name] = local_repos[repo_name]

    if args.repository:
        for name in args.repository:
            add_repo_and_deps(name)
    else:
        for name in local_repos:
            add_repo_and_deps(name)

    sync(to_sync.values())


def add_sync_parser(subparsers, parents):
    parser = subparsers.add_parser("sync", help=l10n("sync-help"), parents=parents)
    parser.add_argument(
        "repository",
        help=l10n("sync-repositories-help"),
        nargs="*",
    )
    parser.set_defaults(func=sync_args)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Various utility functions
"""

from collections import namedtuple
from typing import Any, Iterable, List, Optional, Tuple

from portmod.config.license import has_eula, is_license_accepted
from portmod.repo.keywords import accepts, accepts_testing, get_keywords
from portmodlib.atom import QualifiedAtom, version_gt

from .globals import env
from .pybuild import Pybuild

KeywordDep = namedtuple("KeywordDep", ["atom", "keyword", "masked"])
LicenseDep = namedtuple("LicenseDep", ["atom", "license", "is_eula", "repo"])


def is_keyword_masked(arch: str, keywords: Iterable[str]):
    return "-" + arch in keywords or (
        "-*" in keywords and arch not in keywords and "~" + arch not in keywords
    )


def get_keyword_dep(package: Pybuild) -> Optional[KeywordDep]:
    if not accepts(get_keywords(package.ATOM), package.KEYWORDS):
        arch = env.prefix().ARCH
        if accepts_testing(arch, package.KEYWORDS):
            return KeywordDep(QualifiedAtom(package.ATOM.CP), "~" + arch, False)
        return KeywordDep(
            QualifiedAtom("=" + package.ATOM.CP),
            "**",
            is_keyword_masked(arch, package.KEYWORDS),
        )
    return None


def select_package(packages: Iterable[Pybuild]) -> Tuple[Pybuild, Any]:
    """
    Chooses a mod version based on keywords and accepts it if the license is accepted
    """
    if not packages:
        raise Exception("Cannot select mod from empty modlist")

    filtered = list(
        filter(lambda pkg: accepts(get_keywords(pkg.ATOM), pkg.KEYWORDS), packages)
    )

    keyword = None

    if filtered:
        mod = max(filtered, key=lambda pkg: pkg.version)
    else:
        arch = env.prefix().ARCH
        # No mods were accepted. Choose the best version and add the keyword
        # as a requirement for it to be installed
        unstable = list(
            filter(lambda mod: accepts_testing(arch, mod.KEYWORDS), packages)
        )
        if unstable:
            mod = max(unstable, key=lambda pkg: pkg.version)
            keyword = "~" + arch
        else:
            # There was no mod that would be accepted by enabling testing.
            # Try enabling unstable
            mod = max(packages, key=lambda pkg: pkg.version)
            keyword = "**"

    deps: List[Any] = []
    if not is_license_accepted(mod, mod.get_use()):
        deps.append(LicenseDep(mod.CPN, mod.LICENSE, has_eula(mod), mod.REPO))
    if keyword is not None:
        deps.append(
            KeywordDep(
                QualifiedAtom("=" + mod.ATOM.CPF),
                keyword,
                is_keyword_masked(env.prefix().ARCH, mod.KEYWORDS),
            )
        )

    return (mod, deps or None)


def get_max_version(versions: Iterable[str]) -> Optional[str]:
    """
    Returns the largest version in the given list

    Version should be a valid version according to PMS section 3.2,
    optionally follwed by a revision

    Returns None if the version list is empty

    .. warning::
        Deprecated and will be removed in portmod 2.6
    """
    newest = None
    for version in versions:
        if newest is None or version_gt(version, newest):
            newest = version
    return newest


def get_newest(packages: Iterable[Pybuild]) -> Pybuild:
    """
    Returns the newest mod in the given list based on version

    .. warning::
        Deprecated and will be removed in portmod 2.6
    """
    return max(packages, key=lambda pkg: pkg.version)


def sort_by_version(packages: Iterable[Pybuild]) -> List[Pybuild]:
    """
    Sorts the given packages in order of version

    .. warning::
        Deprecated and will be removed in portmod 2.6
    """
    return sorted(packages, key=lambda pkg: pkg.version)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Helper functions for interacting with repository metadata files.
"""

import glob
import os
from collections import defaultdict
from typing import Dict, Iterable, List, Mapping, Optional, Set, Tuple

from portmod.config import read_config
from portmod.functools import system_cache
from portmod.pybuild import Pybuild
from portmodlib.atom import QualifiedAtom
from portmodlib.parsers.list import read_list
from portmodlib.portmod import (
    GroupDeclaration,
    PackageMetadata,
    parse_category_metadata,
    parse_groups,
    parse_package_metadata,
    parse_yaml_dict,
    parse_yaml_dict_dict,
)

from . import Repo, get_repo

# Note: All repository files should be encoded as utf-8.


def __get_layout(repo_path: str) -> Dict:
    path = os.path.join(repo_path, "metadata", "layout.conf")
    if os.path.exists(path):
        return read_config(path, {})
    return {}


@system_cache
def is_pybuild_version_banned(repo_path: str, version: int) -> bool:
    banned_versions = set(__get_layout(repo_path).get("pybuild_versions_banned", []))
    return version in banned_versions


@system_cache
def is_pybuild_version_deprecated(repo_path: str, version: int) -> bool:
    banned_versions = set(
        __get_layout(repo_path).get("pybuild_versions_deprecated", [])
    )
    return version in banned_versions


def get_master_names(repo_path: str) -> Set[str]:
    """Returns the direct masters for the repository at the given path"""
    masters = set()
    for master in __get_layout(repo_path).get("masters", "").split():
        masters.add(master)
    return masters


@system_cache
def get_masters(repo_path: str) -> List[Repo]:
    """Returns the direct masters for the repository at the given path"""
    return [get_repo(master) for master in get_master_names(repo_path)]


@system_cache
def get_categories(repo: str) -> Set[str]:
    """Retrieves the list of categories given a path to a repo"""
    categories: Set[str] = {"common"}
    path = os.path.join(repo, "profiles", "categories")
    if os.path.exists(path):
        categories |= set(read_list(path, encoding="utf-8"))
    for master in get_masters(repo):
        categories |= get_categories(master.location)

    return categories


@system_cache
def get_archs(repo: str) -> Set[str]:
    """Returns the available architectures in a given repo"""
    archs: Set[str] = set()
    path = os.path.join(repo, "profiles", "arch.list")
    if os.path.exists(path):
        archs |= set(read_list(path, encoding="utf-8"))

    return archs


@system_cache
def get_global_use(repo) -> Dict[str, str]:
    """
    Returns the global use flag declarations for a given repository

    Each mapping in the result dictionary represents a use flag
    and its description
    """
    use: Dict[str, str] = {}
    path = os.path.join(repo, "profiles", "use.yaml")
    if os.path.exists(path):
        use.update(parse_yaml_dict(path))

    for master in get_masters(repo):
        use.update(get_global_use(master.location))

    return use


def get_profiles(
    arch: str, repos: Iterable[Repo]
) -> Dict[str, List[Tuple[str, str, str]]]:
    """
    Returns the list of profiles available from all known repositories

    args:
        arch: Only profiles matching this architecture will be produced
        repos: Only profiles from these repositories will be produced

    returns:
        A tuple containing the profile's path, name stability keyword
        and repository
    """
    profiles: Dict[str, List[Tuple[str, str, str]]] = defaultdict(list)

    for repo in repos:
        path = os.path.join(repo.location, "profiles", "profiles.yaml")
        if os.path.exists(path):
            repo_profiles = parse_yaml_dict_dict(path)
            for profile in sorted(repo_profiles.get(arch, [])):
                path = os.path.join(repo.location, "profiles", profile)
                profiles[repo.name].append(
                    (path, profile, repo_profiles[arch][profile])
                )
    return profiles


@system_cache
def license_exists(repo: str, name: str) -> bool:
    """
    Returns true if the given license name corresponds to a
    licence in the repository
    """
    path = os.path.join(repo, "licenses", name)
    if os.path.exists(path):
        return True

    for master in get_masters(repo):
        if license_exists(master.location, name):
            return True

    return False


@system_cache
def get_license(repo: str, name: str) -> str:
    """Returns the full content of the named license"""
    path = os.path.join(repo, "licenses", name)
    if os.path.exists(path):
        with open(path, mode="r", encoding="utf-8") as license_file:
            return license_file.read()
    else:
        for master in get_masters(repo):
            license_contents = get_license(master.location, name)
            if license is not None:
                return license_contents

        raise Exception(f"Nonexistant license: {name}")


@system_cache
def get_license_groups(repo: str) -> Dict[str, Set[str]]:
    """
    Returns license groups defined by this repository and its masters

    @param repo: path to repository
    @returns set of license groups
    """
    result: Dict[str, Set[str]] = defaultdict(set)

    for master in get_masters(repo):
        # Merge each set into the result. Licenses can only be added to groups
        for group, licenses in get_license_groups(master.location).items():
            result[group] |= licenses

    path = os.path.join(repo, "profiles", "license_groups.yaml")
    if os.path.exists(path):
        groups = parse_yaml_dict(path)

        for name, values in groups.items():
            if values is not None:
                result[name] |= set(values.split())

    def substitute(group: str):
        groups = []
        for license_str in result[group]:
            if license_str.startswith("@"):
                groups.append(license_str)
        for subgroup in groups:
            result[group].remove(subgroup)
            substitute(subgroup.lstrip("@"))
            result[group] |= result[subgroup.lstrip("@")]

    for group in result:
        substitute(group)

    return result


@system_cache
def get_maintainer_groups(repo_path: str) -> Dict[str, GroupDeclaration]:
    """
    Returns license groups defined by this repository and its masters

    @param repo: path to repository
    @returns set of license groups
    """
    result = {}

    for master in get_masters(repo_path):
        result.update(get_maintainer_groups(master.location))

    path = os.path.join(repo_path, "metadata", "groups.yaml")
    if os.path.exists(path):
        result = parse_groups(path)

    return result


@system_cache
def get_package_metadata(mod: Pybuild) -> Optional[PackageMetadata]:
    """Loads the metadata file for the given mod"""
    path = os.path.join(os.path.dirname(mod.FILE), "metadata.yaml")
    if not os.path.exists(path):
        return None

    return parse_package_metadata(path)


@system_cache
def get_category_metadata(repo: str, category: str):
    """Loads the metadata file for the given category"""
    path = os.path.join(repo, category, "metadata.yaml")

    if os.path.exists(path):
        return parse_category_metadata(path)

    for master in get_masters(repo):
        metadata = get_category_metadata(master.location, category)
        if metadata is not None:
            return metadata

    return None


@system_cache
def get_use_expand(repo: str) -> Set[str]:
    """Returns all possible use expand values for the given repository"""
    groups = set()
    for file in glob.glob(os.path.join(repo, "profiles", "desc", "*.yaml")):
        use_expand, _ = os.path.splitext(os.path.basename(file))
        groups.add(use_expand.upper())
    for master in get_masters(repo):
        groups |= get_use_expand(master.location)

    return groups


@system_cache
def get_use_expand_values(repo: str, use_expand: str) -> Dict[str, str]:
    """Returns all possible use expand values for the given repository"""
    values = {}
    for master in get_masters(repo):
        values.update(get_use_expand_values(master.location, use_expand))

    lowered = use_expand.lower()
    path = os.path.join(repo, "profiles", "desc", lowered + ".yaml")
    if os.path.exists(path):
        kvps = parse_yaml_dict(path)
        values.update(kvps)

    return values


@system_cache
def check_use_expand_flag(repo: str, variable: str, flag: str) -> bool:
    """
    Returns true if the given use flag is declared
    in a USE_EXPAND desc file for the given variable
    """
    path = os.path.join(repo, "profiles", "desc", variable.lower() + ".yaml")
    if os.path.exists(path):
        if flag in parse_yaml_dict(path):
            return True

    for master in get_masters(repo):
        if check_use_expand_flag(master.location, variable, flag):
            return True

    return False


@system_cache
def get_use_flag_atom_aliases(repo_path: str) -> Mapping[str, QualifiedAtom]:
    """Returns aliases between use flags and atoms"""
    path = os.path.join(repo_path, "profiles", "use.alias.yaml")
    mapping = {}
    if os.path.exists(path):
        mapping = parse_yaml_dict(path)

    result: Dict[str, str] = {}

    for master in get_masters(repo_path):
        master_mapping = get_use_flag_atom_aliases(master.location)
        result.update(master_mapping)
        for key in master_mapping:
            if key in mapping and master_mapping[key] != mapping[key]:
                raise RuntimeError(
                    f"Alias {key}={mapping[key]} overrides alias {key}={master_mapping[key]} "
                    "defined by master repository {master.name}"
                )

    result.update(mapping)
    return {flag: QualifiedAtom(atom) for flag, atom in result.items()}



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
CLI for interacting with individual pybuild files
"""

import argparse
import json
import logging
import lzma
import os
import pickle
import re
import sys
from traceback import extract_tb, format_exception_only, format_list
from types import SimpleNamespace

from ._loader import load_file, load_installed
from ._phase import PhaseState
from .execute import execute
from .fs import is_parent
from .globals import prefix_name, vdb_path
from .log import init_logger
from .pybuild import FullPybuild
from .util import pybuild_dumper


def excepthook(typ, value, tb):
    """An exception hook which suppresses frames from this module in the traceback"""
    if value.__cause__:
        cause = value.__cause__
        excepthook(type(cause), cause, cause.__traceback__)
        print()
        print(
            "The above exception was the direct cause of the following exception:"
            + os.linesep
        )
    if value.__context__:
        context = value.__context__
        excepthook(type(context), context, context.__traceback__)
        print()
        print(
            "During handling of the above exception, another exception occurred:"
            + os.linesep
        )
    show = [
        fs
        for fs in extract_tb(tb)
        if fs.filename and fs.filename not in (__file__, "<string>")
    ]
    fmt = format_list(show) + format_exception_only(typ, value)
    print("Traceback (most recent call last):")
    print("".join(fmt), end="", file=sys.stderr)


def wrapper_pybuild(args):
    # Set info variables
    from pybuild.info import _set_info

    _set_info(args.pybuild_file)

    pkg: FullPybuild
    if prefix_name() and is_parent(
        os.path.abspath(os.path.normpath(args.pybuild_file)), vdb_path()
    ):
        pkg = load_installed(args.pybuild_file)
    else:
        pkg = load_file(args.pybuild_file)

    if args.state_dir and os.path.exists(args.state_dir):
        state_file = os.path.join(args.state_dir, "state.pickle")
        if os.path.exists(state_file):
            with open(state_file, "rb") as b_file:
                pkg.__dict__ = pickle.load(b_file)

    if args.initial_state:
        state = PhaseState.from_json(json.loads(args.initial_state))
        pkg.__dict__.update(state.__dict__)

    def pkg_func(pkg, name):
        pkg.execute = execute
        func = getattr(pkg, name)
        return func()

    for command in args.command:
        if command == "unpack":
            pkg_func(pkg, "src_unpack")
        elif command == "prepare":
            pkg_func(pkg, "src_prepare")
        elif command == "install":
            pkg_func(pkg, "src_install")
        elif command == "postinst":
            pkg_func(pkg, "pkg_postinst")
        elif command == "pretend":
            pkg_func(pkg, "pkg_pretend")
        elif command == "nofetch":
            pkg_func(pkg, "pkg_nofetch")
        elif command == "can-update-live":
            result = pkg_func(pkg, "can_update_live")
            if not isinstance(result, bool):
                raise RuntimeError(
                    f"can_update_live returned unexpected result {result}"
                )
            sys.exit(int(result) * 142)

    if args.state_dir:
        with open(os.path.join(args.state_dir, "state.pickle"), "wb") as b_file_out:
            pickle.dump(pkg.__dict__, b_file_out)

        with open(os.path.join(args.state_dir, "environment.xz"), "wb") as b_file_out:
            # Keys are sorted to produce consistent results and
            # easy to read commits in the DB
            dictionary = pkg.__class__.__dict__.copy()
            dictionary.update(pkg.__dict__)
            dictionary = dict(
                filter(
                    lambda elem: not elem[0].startswith("_") and elem[0] != "execute",
                    dictionary.items(),
                )
            )
            jsonstr = json.dumps(dictionary, default=pybuild_dumper, sort_keys=True)
            b_file_out.write(lzma.compress(str.encode(jsonstr)))


def wrapper_module(args):
    if args.initial_state:
        state = PhaseState.from_json(json.loads(args.initial_state))

    name, _ = os.path.splitext(os.path.basename(args.module_file))
    module = os.path.basename(os.path.dirname(args.module_file))
    sys.path.append(os.path.dirname(os.path.dirname(args.module_file)))
    with open(args.module_file, "r", encoding="utf-8") as file:
        tmp_globals = {
            "__builtins__": __builtins__,
            "__name__": name,
            "__package__": module,
        }
        code = compile(file.read(), args.module_file, "exec")
        exec(code, tmp_globals)

    do_functions = {}
    for globname in tmp_globals:
        if globname.startswith("do_"):
            name = re.sub("^do_", "", globname)
            do_functions[name] = tmp_globals[globname]

    assert args.command == "execute" and args.module_func is not None
    if args.module_func == "update":
        do_functions[args.module_func](state)
    else:
        function_args = None
        if args.args:
            function_args = SimpleNamespace(**json.loads(args.args))
        do_functions[args.module_func](state, function_args)


def main():
    """
    Wrapper script for directly loading pybuild and module files.
    This should always be invoked using the executable sandbox,
    and is not intended to be invoked manually.
    """
    parser = argparse.ArgumentParser(description=main.__doc__)
    parser.add_argument("--verbosity", help="verbosity level")
    subparsers = parser.add_subparsers()
    pybuild = subparsers.add_parser("pybuild", help="interact with pybuild files")
    module = subparsers.add_parser("module", help="interact with module files")
    pybuild.add_argument("pybuild_file", metavar="<pybuild file>")
    pybuild.add_argument("command", metavar="<command>", nargs="+")
    pybuild.add_argument(
        "--state-dir",
        help="The path of a directory to be used to store state information",
    )
    pybuild.add_argument(
        "--initial-state",
        help="A json-encoded dictionary of values to be set as package attributes",
    )

    module.add_argument("module_file", metavar="<module file>")
    module.add_argument("command", metavar="<command>")
    module.add_argument(
        "--module-func", help="name of function to execute if command is execute"
    )
    module.add_argument("--args", help="arguments to pass to the function")
    module.add_argument(
        "--initial-state",
        help="A json-encoded dictionary of values to be set as package attributes",
    )

    pybuild.set_defaults(func=wrapper_pybuild)
    module.set_defaults(func=wrapper_module)

    args = parser.parse_args()

    if len(sys.argv) == 1:
        parser.print_help()

    args.verbose = None
    args.quiet = None
    init_logger(args)
    # Set logging level manually. Args doesn't contain any of --verbose or --quiet
    if args.verbosity is not None:
        logging.root.setLevel(args.verbosity)

    # In debug-level verbosity or lower, use default full traces
    # Otherwise suppress the parts referencing this module
    if logging.root.level > logging.DEBUG:
        sys.excepthook = excepthook
    args.func(args)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Functions to set up and tear down a testing environment
"""

import os
import shutil
import sys
from locale import LC_ALL, setlocale
from logging import error
from tempfile import mkdtemp
from types import SimpleNamespace
from typing import Any, Dict, Optional

from portmod.cache import cache
from portmod.config import set_config_value
from portmod.config.profiles import profile_parents
from portmod.functools import clear_install_cache, clear_system_cache
from portmod.globals import env, refresh_env
from portmod.prefix import add_prefix, get_prefixes
from portmod.repos import Repo
from portmod.sync import sync
from portmod.vdb import VDB
from portmodlib.fs import onerror
from portmodlib.log import init_logger

TEST_REPO_DIR = os.path.join(os.path.dirname(__file__), "testrepo")
TEST_REPO = Repo("test", TEST_REPO_DIR, priority=-1000)
_TMP_FUNC = None
TESTDIR: Optional[str] = None
OLD: Optional[Dict[str, Any]] = None
OLD_CWD: Optional[str] = None


def set_test_repo():
    """Replaces the repo list with one that just contains the test repo"""
    os.makedirs(os.path.dirname(env.REPOS_FILE), exist_ok=True)
    with open(env.REPOS_FILE, "w") as file:
        print("[test]", file=file)
        print(f"location = {TEST_REPO.location}", file=file)
        print("auto_sync = False", file=file)


def setup_no_prefix():
    global OLD, OLD_CWD, TESTDIR
    # Use C locale. This will fail to read files containing unicode,
    # unless the files are supposed to and we explicitly open them as utf-8
    setlocale(LC_ALL, None)
    init_logger(SimpleNamespace(verbose=False, quiet=False))

    cwd = os.getcwd()
    clear_system_cache()
    clear_install_cache()
    OLD = env.__dict__
    OLD_CWD = cwd
    TESTDIR = mkdtemp(prefix="portmod.test")
    env.CONFIG_DIR = os.path.join(TESTDIR, "config")
    env.CACHE_DIR = os.path.join(TESTDIR, "cache")
    env.DATA_DIR = os.path.join(TESTDIR, "local")
    env.INTERACTIVE = False
    env.TESTING = True
    env.DEBUG = True
    env.PREFIX_NAME = None

    os.makedirs(TESTDIR, exist_ok=True)
    os.makedirs(os.path.join(TESTDIR, "local"), exist_ok=True)
    os.makedirs(os.path.join(TESTDIR, "work"), exist_ok=True)
    os.chdir(os.path.join(TESTDIR, "work"))

    refresh_env()
    set_test_repo()
    refresh_env()


def setup_env(profile, directory: Optional[str] = None):
    """
    Sets up an entire testing environment
    All file writes will occur within a temporary directory as a result
    """
    global TESTDIR
    setup_no_prefix()
    assert TESTDIR
    if "test" not in get_prefixes():
        add_prefix("test", "test", directory)
    env.set_prefix("test")
    select_profile(profile)
    refresh_env()
    if env.PREFIX_NAME is not None:
        set_config_value("REPOS", "test")

    with VDB() as gitrepo:
        gitrepo.config_writer().set_value("commit", "gpgsign", False).release()
        gitrepo.config_writer().set_value(
            "user", "email", "pytest@example.com"
        ).release()
        gitrepo.config_writer().set_value("user", "name", "pytest").release()

    sync([TEST_REPO])
    return {
        "testdir": TESTDIR,
        "config": f"{TESTDIR}/config.cfg",
        "config_ini": f"{TESTDIR}/config.ini",
    }


def rmtree(path: str):
    """Custom rmtree wrapper to deal with windows problems"""
    if sys.platform == "win32":
        from time import sleep

        iters = 0
        while os.path.exists(path) and iters < 10:
            iters += 1
            try:
                shutil.rmtree(path, onerror=onerror, ignore_errors=True)
            except PermissionError as e:
                error(e)
                sleep(0.01)
    else:
        shutil.rmtree(path, onerror=onerror)


def tear_down_env():
    """
    Reverts env to original state
    """
    assert OLD_CWD and TESTDIR and OLD
    os.chdir(OLD_CWD)
    env.__dict__ = OLD
    cache.clear()
    if os.path.exists(TESTDIR):
        rmtree(TESTDIR)


def unset_profile():
    """Removes the profile link"""
    linkpath = os.path.join(env.prefix().CONFIG_DIR, "profile")
    if os.path.exists(linkpath):
        os.unlink(linkpath)
    profile_parents.cache_clear()


def select_profile(profile):
    """Selects the given test repo profile"""
    linkpath = os.path.join(env.prefix().CONFIG_DIR, "profile")
    unset_profile()
    os.symlink(os.path.join(TEST_REPO_DIR, "profiles", profile), linkpath)
    profile_parents.cache_clear()



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

from __future__ import annotations

from abc import ABC, abstractmethod
from collections import defaultdict
from logging import warning
from types import SimpleNamespace
from typing import Any, Dict, Iterable, List, Optional, Sequence, Set, Tuple, Union

from pysat.formula import WCNFPlus

from portmod.config.mask import get_masked, get_unmasked
from portmod.config.use import get_forced_use
from portmod.loader import load_pkg, load_pkg_fq
from portmod.pybuild import Pybuild
from portmodlib.atom import Atom, FQAtom, QualifiedAtom, atom_sat
from portmodlib.colour import blue, green
from portmodlib.usestr import human_readable_required_use, parse_usestr

from .tokens import (
    AtomToken,
    FlagToken,
    Token,
    VariableToken,
    expand_use_conditionals,
    token_conflicts,
)


class Target(SimpleNamespace):
    def __init__(self, pkgs: List[Pybuild], atom: Atom, source: Optional[str]):
        self.pkgs = pkgs
        self.atom = atom
        self.source = source


class Formula:
    """
    Intermediate representation of the integer WCNFPlus SAT formula accepted by pysat

    All flags are disabled by prepending a "-"

    Atom requirements are represented by the fully qualified atom,
    as produced by mod.ATOM

    Use flag requirements are represented by the fully qualified atom
    of the mod they apply to, followed by [flag]

    Custom variables are prefixed by an underscore and are
    used only for the calculation and ignored in the output
    """

    __i = 1  # Integer counter, for the numerical equivalent of tokens
    __j = 1  # Variable name counter.
    __numcache: Dict[Token, int] = {}  # Strings are
    __stringcache: Dict[int, Token] = {}
    __variabledesc: Dict[Token, str] = {}

    class Clause(ABC):
        """Generic clause type"""

        def __init__(self, clause: Iterable[Token], source):
            self.requirements: Set[Token] = set()
            self.clause = clause
            self.intclause: Optional[Iterable[int]] = None
            self.source = source

        def str2num(self):
            """
            Converts the tokens in the clause to integers for use with pysat
            """
            self.intclause = list(map(Formula.getnum, self.clause))
            return self

        @abstractmethod
        def colourless(self) -> str:
            ...

        def blocks(self, _model: Set[Token], _clause: Formula.Clause) -> bool:
            return False

    class MetaClause(Clause):
        def __init__(
            self,
            source: Optional[str],
            desc: Optional[str],
            clause: Iterable[Token],
            weight: Optional[int],
            atmost: Optional[int],
        ):
            self.source = source
            self.desc = desc
            self.clause = clause
            self.weight = weight
            self.atmost = atmost
            self.requirements: Set[Token] = set()

        def colourless(self) -> str:
            return str(self)

        def __str__(self):
            return f"{self.source} - {self.desc}"

    class DepClause(Clause):
        def __init__(self, source: str, clause: Iterable[Token], dependency: Atom):
            super().__init__(clause, source)
            self.dependency = dependency

        def __str__(self):
            return f"{green(self.dependency)}: required by {green(self.source)}"

        def colourless(self):
            return f"{self.dependency}: required by {self.source}"

        def blocks(self, model: Set[Token], clause: Formula.Clause) -> bool:
            if isinstance(clause, Formula.BlockerClause):
                if atom_sat(self.dependency, clause.blocked) and all(
                    req in model for req in self.requirements
                ):
                    return True

            return False

    class BlockerClause(Clause):
        def __init__(self, source: str, clause: Iterable[Token], blocked: Atom):
            super().__init__(clause, source)
            self.blocked = blocked

        def __str__(self):
            return f"{green(self.blocked)}: blocked by {green(self.source)}"

        def colourless(self):
            return f"{self.blocked}: blocked by {self.source}"

        def blocks(self, model: Set[Token], clause: Formula.Clause) -> bool:
            if isinstance(clause, Formula.DepClause):
                if atom_sat(self.blocked, clause.dependency) and all(
                    req in model for req in self.requirements
                ):
                    return True

            return False

    class UseDepClause(Clause):
        def __init__(
            self, source: str, clause: Iterable[Token], depatom: Atom, flag: FlagToken
        ):
            super().__init__(clause, source)
            self.depatom = depatom
            self.flag = flag

        def __str__(self):
            return (
                f"{green(self.depatom)}[{self.flag}]: required by {green(self.source)}"
            )

        def colourless(self):
            return f"{self.depatom}[{self.flag}]: required by {self.source}"

        def blocks(self, model: Set[Token], clause: Formula.Clause) -> bool:
            if isinstance(clause, Formula.UseDepClause):
                if (
                    atom_sat(self.depatom, clause.depatom)
                    and token_conflicts(self.flag, clause.flag)
                    and all(req in model for req in self.requirements)
                ):
                    return True
            elif isinstance(clause, Formula.RequiredUseClause):
                if (
                    atom_sat(clause.atom, self.depatom)
                    and token_conflicts(self.flag, clause.flag)
                    and all(req in model for req in self.requirements)
                ):
                    return True

            return False

    class RequiredUseClause(Clause):
        def __init__(
            self,
            atom: Atom,
            clause: Iterable[Token],
            flag: Token,
            tokens: List[Union[str, List]],
        ):
            super().__init__(clause, atom)
            self.atom = atom
            self.flag = flag
            self.tokens = tokens

        def _get_required_use_str(self) -> Tuple[str, str]:
            if isinstance(self.flag, VariableToken):  # If flag is a generated variable
                string = Formula.get_variable_desc(self.flag)
            elif isinstance(self.flag, FlagToken):
                if self.flag.polarity:
                    string = list(self.flag.value.USE)[0]
                else:
                    string = "-" + list(self.flag.value.USE)[0]
            else:
                raise Exception(
                    f"AtomToken {self.flag} passed as the flag to RequiredUseClause!"
                )
            parent = human_readable_required_use(self.tokens)
            return string, parent

        def __str__(self):
            string, parent = self._get_required_use_str()
            # FIXME: Localize
            if string == parent:
                return f"{green(self.atom)} could not satisfy {blue(string)}"
            else:
                return (
                    f"{green(self.atom)} could not satisfy {blue(string)}, which is part of "
                    f"the larger clause {blue(parent)}"
                )

        def colourless(self):
            string, parent = self._get_required_use_str()
            if string == parent:
                return f"{self.atom} could not satisfy {string}"
            else:
                return (
                    f"{self.atom} could not satisfy {string}, which is part of "
                    f"the larger clause {parent}"
                )

        def blocks(self, model: Set[Token], clause: Formula.Clause) -> bool:
            if isinstance(clause, Formula.UseDepClause):
                if (
                    atom_sat(self.atom, clause.depatom)
                    and token_conflicts(self.flag, clause.flag)
                    and all(req in model for req in self.requirements)
                ):
                    return True
            elif isinstance(clause, Formula.RequiredUseClause):
                if (
                    atom_sat(self.atom, clause.atom)
                    and token_conflicts(self.flag, clause.flag)
                    and all(req in model for req in self.requirements)
                ):
                    return True

            return False

    def __init__(self):
        self.clauses = []
        self.atoms: Dict[QualifiedAtom, Set[FQAtom]] = defaultdict(set)
        self.flags: Set[FQAtom] = set()

    @classmethod
    def getnum(cls, token: Token) -> int:
        if token in cls.__numcache:
            return cls.__numcache[token]

        if token.polarity:
            cls.__numcache[token] = cls.__i
            cls.__numcache[token.neg()] = -cls.__i
            cls.__stringcache[cls.__i] = token
            cls.__stringcache[-cls.__i] = token.neg()
        else:
            cls.__numcache[token] = -cls.__i
            cls.__numcache[token.neg()] = cls.__i
            cls.__stringcache[-cls.__i] = token
            cls.__stringcache[cls.__i] = token.neg()

        cls.__i += 1
        return cls.__numcache[token]

    @classmethod
    def getstring(cls, num: int) -> Token:
        return cls.__stringcache[num]

    @classmethod
    def genvariable(cls, desc: List[Any]) -> Token:
        var = VariableToken("__" + str(cls.__j))
        cls.__j += 1
        cls.__variabledesc[var] = human_readable_required_use(desc)
        return var

    @classmethod
    def get_variable_desc(cls, var: Token) -> str:
        return cls.__variabledesc[var]

    def merge(self, other: Formula):
        self.clauses.extend(other.clauses)
        self.flags |= other.flags
        for atom, values in other.atoms.items():
            self.atoms[atom] |= values
        return self

    def get_wcnfplus(self) -> WCNFPlus:
        formula = WCNFPlus()
        for clause in self.clauses:
            if isinstance(clause, Formula.MetaClause) and clause.weight is not None:
                formula.append(clause.intclause, weight=clause.weight)
            elif isinstance(clause, Formula.MetaClause) and clause.atmost is not None:
                formula.append([clause.intclause, clause.atmost], is_atmost=True)
            else:
                formula.append(clause.intclause)
        return formula

    def append(
        self,
        clause: Iterable[Token],
        from_atom: Optional[str] = None,
        desc: Optional[str] = None,
        weight=None,
        atmost=None,
    ):
        self.clauses.append(Formula.MetaClause(from_atom, desc, clause, weight, atmost))
        self.__update_for_clause__(clause)

    def append_dep(self, clause: Iterable[Token], from_atom: str, dependency: Atom):
        self.clauses.append(Formula.DepClause(from_atom, clause, dependency))
        self.__update_for_clause__(clause)

    def append_blocker(self, clause: Iterable[Token], from_atom: str, blocked: Atom):
        self.clauses.append(Formula.BlockerClause(from_atom, clause, blocked))
        self.__update_for_clause__(clause)

    def append_required_use(
        self,
        clause: Iterable[Token],
        from_atom: Atom,
        flag: Token,
        tokens: List[Union[str, List]],
    ):
        self.clauses.append(Formula.RequiredUseClause(from_atom, clause, flag, tokens))
        self.__update_for_clause__(clause)

    def append_usedep(
        self, clause: Iterable[Token], from_atom: str, dep_atom: Atom, flag: FlagToken
    ):
        self.clauses.append(Formula.UseDepClause(from_atom, clause, dep_atom, flag))
        self.__update_for_clause__(clause)

    def __update_for_clause__(self, clause: Iterable[Token]):
        for token in clause:
            if not isinstance(token, VariableToken):
                if isinstance(token, FlagToken):
                    self.flags.add(token.value)
                elif isinstance(token, AtomToken):
                    self.atoms[QualifiedAtom(token.value.CPN)].add(token.value)

    def extend(
        self,
        from_atom: Atom,
        clauses: List[List[Token]],
        desc: Optional[str] = None,
        weight=None,
        atmost=None,
    ):
        for clause in clauses:
            self.append(clause, from_atom, desc, weight, atmost)

    def add_constraints(self, constraints: List[Token]):
        self.__update_for_clause__(constraints)
        for clause in self.clauses:
            if not (
                isinstance(clause, Formula.MetaClause) and clause.atmost is not None
            ):
                clause.clause = list(clause.clause) + constraints
                clause.requirements |= {token.neg() for token in constraints}

    def make_numeric(self):
        for clause in self.clauses:
            clause.str2num()


def get_atmost_one_formulae(tokens: Sequence[Token]) -> List[List[Token]]:
    """
    Returns a list of clauses that enforce that at most one of the tokens may be true

    Note that this can also be achieved by using Formula.append with atmost set to 1,
    however  this does not provide a mechanism for handling additional conditions.
    Instead, you can use this function, and add the condition to each clause it produces
    """
    if len(tokens) <= 1:
        return []

    result = []
    # Enforce that for any two tokens in the list, one must be false
    for token in tokens[1:]:
        # Invert value of firsttoken
        firsttoken = tokens[0].neg()

        # Invert value of the other token
        othertoken = token.neg()
        result.append([firsttoken, othertoken])

    return result + get_atmost_one_formulae(tokens[1:])


def toggle_clause(tokens: List[Token]) -> List[Token]:
    newtokens = []
    for token in tokens:
        newtokens.append(token.neg())
    return newtokens


def get_required_use_formula(
    mod: Pybuild, tokens: List[Union[str, List]], use_expand: Optional[str] = None
) -> Formula:
    """
    Adds clauses to the given formula for the given mod's REQUIRED_USE

    :param tokens: List of tokens corresponding to the REQUIRED_USE string, parsed
            beforehand by parse_usestr to be a list of tokens, with sublists
            corresponding to brackets in the original string
    """

    def get_required_use_formula_inner(
        tokens: List[Union[str, List]]
    ) -> Tuple[Formula, List[Token]]:
        formula = Formula()
        clausevars = []

        for token in tokens:
            if isinstance(token, list):
                if token[0] != "??" and token[0].endswith("?"):
                    newvar = Formula.genvariable([token])
                    subformulae, subvars = get_required_use_formula_inner(token[1:])
                    usedep = cond_flagstr(mod.ATOM, token[0].rstrip("?")).neg()
                    subformulae.add_constraints([usedep, newvar.neg()])
                    # for clause in get_atmost_one_formulae(subvars):
                    #    formula.append_required_use(
                    #        ["-" + newvar] + clause, mod.ATOM, token
                    #    )
                    formula.merge(subformulae)
                    # Generated variable is added to clausevars, and is free if
                    # condition is unsatisfied, and matches subformulae if condition
                    # is satisfied
                    clausevars.append(newvar)
                else:
                    subformulae, subvars = get_required_use_formula_inner(token[1:])
                    newvar = Formula.genvariable([token])
                    # Note: newvar will only have the value False if the formula
                    # is satisfied
                    if token[0] in ("??", "^^"):
                        for clause in get_atmost_one_formulae(subvars):
                            formula.append(
                                [newvar.neg()] + clause,
                                mod.ATOM,
                                human_readable_required_use(tokens),
                            )
                    if token[0] in ("||", "^^"):
                        formula.append(
                            [newvar.neg()] + subvars,
                            mod.ATOM,
                            human_readable_required_use(tokens),
                        )
                    if token[0] in ("||", "^^", "??"):
                        # If clause is satisfied, and the operator is not AND,
                        # then the subclauses don't need to be all satisfied
                        subformulae.add_constraints([newvar])
                    formula.merge(subformulae)
                    clausevars.append(newvar)

            else:
                flag = token.lstrip("!")
                if use_expand:
                    flag = use_expand + "_" + flag
                var = fstr(mod.ATOM, flag)
                if token.startswith("!"):
                    var = var.neg()

                formula.append_required_use([var], mod.ATOM, var, tokens)
                clausevars.append(var)
        return formula, clausevars

    formula, clausevars = get_required_use_formula_inner(tokens)
    # Top level is an and, so require that all returned variables are satisfied
    for var in clausevars:
        formula.append_required_use([var], mod.ATOM, var, tokens)
    return formula


def fstr(atom: FQAtom, flag: str) -> FlagToken:
    """
    Produces a flag token for the formula given an atom and a flag

    This function does not produce disabled tokens. If a disabled token is
    desired, the result should be negated.
    """
    assert flag[0] not in ("-", "!") and flag[-1] not in ("?", "=")
    return FlagToken(atom.use(flag))


def cond_flagstr(atom: FQAtom, flag: str, negtoken: str = "!") -> FlagToken:
    """
    Given an atom and a flag from a use conditional,
    produces a token for use in the dependency formula
    """
    disabled = flag.startswith(negtoken)
    flag = flag.lstrip(negtoken)
    if disabled:
        return fstr(atom, flag).neg()
    return fstr(atom, flag)


def get_dep_formula(mod: Pybuild, tokens) -> Tuple[Formula, Set[FQAtom]]:
    """
    Adds clauses to the given formula for the dependency strings of the given mod

    :param tokens: List of tokens corresponding to the dependency string, parsed
            beforehand by parse_usestr to be a list of tokens, with sublists
            corresponding to brackets in the original string
    """

    formula = Formula()
    deps: Set[FQAtom] = set()

    for token in expand_use_conditionals(tokens):
        if isinstance(token, list):
            if token[0] == "||":
                # If token is an or, next token is a list, at least one of the elements of
                # which must be satisfied
                orvars = []

                # Create clause for each part of the || expression.
                for subclause in token[1:]:
                    # Create new variable to represent clause
                    var = Formula.genvariable([token])
                    orvars.append(var)
                    # Either one part of the or must be true, or the variable for
                    # the clause should be false
                    new_formula, new_deps = get_dep_formula(mod, [subclause])
                    new_formula.add_constraints([var.neg()])
                    formula.merge(new_formula)
                    deps |= new_deps

                # We should be able to set at least one of the new variables we
                # introduced to be true, meaning that some other part of their clause
                # must be satisfied
                formula.append(orvars, mod.ATOM, human_readable_required_use([token]))
            elif token[0].endswith("?"):
                new_formula, new_deps = get_dep_formula(mod, token[1:])

                # Note: If flag was disabled, we want the flag enabled in the clause, as it
                # should either be enabled,
                # or some other part of the clause must be true if disabled
                # If flag was enabled, we produce the clause (flag => dep),
                # which is equivalent to (-flag | dep)
                new_formula.add_constraints(
                    [cond_flagstr(mod.ATOM, token[0].rstrip("?")).neg()]
                )
                formula.merge(new_formula)
                deps |= new_deps
            else:
                raise Exception(
                    f"Internal Error: dependency structure {tokens} is invalid"
                )
        # Handle regular dependencies
        else:
            blocker = token.startswith("!!")
            atom = Atom(token.lstrip("!"))

            # Note that load_pkg will only return mods that completely match atom
            # I.e. it will handle any versioned atoms itself
            specificpkgs = load_pkg(atom)
            specificatoms = [m.ATOM for m in load_pkg(atom)]

            if not blocker:
                deps |= set(specificatoms)

            # !!foo[A,B] is equivalent to
            # || ( !!foo foo[-A] foo[-B] )
            # What about !!foo[A?,B]
            # A? ( || ( !!foo foo[-A] foo[-B] ) || ( !!foo foo[-B] ) )

            # At least one specific version of this mod must be enabled
            if blocker and not atom.USE:
                for specatom in specificatoms:
                    formula.append_blocker(
                        [AtomToken(specatom, polarity=False)],
                        mod.ATOM,
                        Atom(token.lstrip("!")),
                    )
            elif not blocker:
                formula.append_dep(
                    [AtomToken(atom) for atom in specificatoms], mod.ATOM, atom
                )

            # For each use flag dependency, add a requirement that the flag must be set
            # This depends on the operators used on the flag. See PMS 8.2.6.4
            for flag in atom.USE:
                for spec in specificpkgs:
                    # Either specific version should not be installed,
                    # or flag must be set (depending on flag operators)

                    # Use requirement is unnecessary unless this specific version
                    # of the mod is enabled
                    new_formula = Formula()
                    if flag.lstrip("-") not in spec.IUSE_EFFECTIVE:
                        continue

                    # 2-style
                    if blocker:
                        flagdep = cond_flagstr(spec.ATOM, flag, negtoken="-").neg()
                        new_formula.append_usedep(
                            [flagdep],
                            mod.ATOM,
                            atom.strip_use(),
                            flagdep,
                        )
                    else:
                        flagdep = cond_flagstr(spec.ATOM, flag, negtoken="-")
                        new_formula.append_usedep(
                            [flagdep],
                            mod.ATOM,
                            atom.strip_use(),
                            flagdep,
                        )

                    new_formula.add_constraints([AtomToken(spec.ATOM, polarity=False)])
                    formula.merge(new_formula)
    return formula, deps


def generate_formula(
    mods: Iterable[Target], depsadded: Set[FQAtom], deep: bool
) -> Formula:
    """
    Generates a hard dependency formula for the given mods

    :param mods: Each entry should contain a list of mods with the same base
                 category and name, the atom that pulled those mods in, and a
                 string describing where the mods were pulled from.
    :param depsadded: Mods that have already been included in the formula and
                      should not be added again
    :returns: The resulting formula
    """
    formula = Formula()
    # Queue of mods to add to the formula
    new: List[Pybuild] = []
    # Ensure newselected and oldselected mods are satisfied
    for target in mods:
        if target.source:
            # If a source is specified, at least one version of the mod must be
            # installed
            # Otherwise, we only include it for the purpose of dependency resolution
            formula.append_dep(
                [AtomToken(mod.ATOM) for mod in target.pkgs], target.source, target.atom
            )
            if target.atom.USE:
                for flag in target.atom.USE:
                    for mod in target.pkgs:
                        flagdep = cond_flagstr(mod.ATOM, flag, negtoken="-")
                        formula.append_usedep(
                            [
                                AtomToken(mod.ATOM, polarity=False),
                                flagdep,
                            ],
                            target.source,
                            target.atom.strip_use(),
                            flagdep,
                        )

        new += target.pkgs

    while new:
        # Mods to parse in next iteration, mapped to the mod that depends on them
        nextmods: Set[FQAtom] = set()
        for mod in new:
            # Either mod must not be installed, or mods dependencies must be satisfied
            new_formula, deps = get_dep_formula(
                mod, parse_usestr(mod.DEPEND + " " + mod.RDEPEND, token_class=Atom)
            )
            new_formula.merge(
                get_required_use_formula(mod, parse_usestr(mod.REQUIRED_USE))
            )
            # Exactly one texture_size flag must be enabled
            if mod.TEXTURE_SIZES.strip():
                new_formula.merge(
                    get_required_use_formula(
                        mod,
                        [["^^"] + parse_usestr(mod.TEXTURE_SIZES)],
                        use_expand="texture_size",
                    )
                )
            new_formula.add_constraints([AtomToken(mod.ATOM, polarity=False)])
            formula.merge(new_formula)
            for flag in get_forced_use(mod.ATOM):
                if flag.lstrip("-") not in mod.IUSE_EFFECTIVE:
                    continue

                formula.append(
                    [cond_flagstr(mod.ATOM, flag, negtoken="-")],
                    "profile use.force or mod.use.force",
                    f"Flag {flag} is forced on mod {mod.ATOM}",
                )

            depsadded.add(mod.ATOM)
            if deep:
                # Add this mod's dependencies to the next set of mods to parse
                nextmods |= deps

        new = []
        for atom in nextmods:
            if atom not in depsadded:
                new.append(load_pkg_fq(atom))

    return formula


def is_unmasked(atom: FQAtom) -> bool:
    for unmasked_atom in get_unmasked().get(atom.CPN, []):
        if atom_sat(atom, unmasked_atom):
            return True
    return False


def add_masked(atoms: Dict[QualifiedAtom, Set[FQAtom]], flags: Iterable[FQAtom]):
    formula = Formula()
    masked = get_masked()
    for cpn in atoms:
        if cpn in masked:
            for atom in atoms[cpn]:
                for masked_atom, comment in masked[cpn].items():
                    if atom_sat(atom, masked_atom) and not is_unmasked(atom):
                        if comment:
                            description = (
                                "package.mask with the following comment:\n  "
                                + blue("\n  ".join(comment))
                            )
                        else:
                            description = "package.mask"
                        # Masks don't apply to installed packages
                        # (i.e. the user should remove them manually)
                        if atom.R.endswith("::installed"):
                            warning(
                                f"Installed package {atom.CPF} is masked by {description}"
                            )
                        else:
                            formula.append_blocker(
                                [AtomToken(atom, polarity=False)],
                                description,
                                masked_atom,
                            )

    return formula
